# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from prometheus_client import Gauge, start_http_server
import re
import logging

logger = logging.getLogger()


# pip install prometheus_client
# Docs at https://github.com/prometheus/client_python
class prometheus(object):
    def __init__(self, config, appname):
        try:
            port = int(config['prometheus'][appname]['server_port'])
        except (ValueError, KeyError):
            logger.warning('prometheus server_port not present in config. running without metrics.')
            self.enable_metrics = False
            return

        self.gauges = {}

        # per docs, app name in metric prefix needs to be one word
        self.appname = re.sub('[^a-zA-Z0-9]+', '', appname)

        logger.info('Starting prometheus metrics web server at %s', port)
        start_http_server(port)
        self.enable_metrics = True

    def send_metrics(self, metrics):
        if not self.enable_metrics:
            return
        for metric, value in metrics.items():
            if metric not in self.gauges:
                self.gauges[metric] = Gauge(self.appname + '_' + metric, '')
            self.gauges[metric].set_to_current_time()
            self.gauges[metric].set(value)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import ujson
import logging
import requests
import random
import time
from gevent import sleep
from iris.constants import SLACK_SUPPORT

logger = logging.getLogger(__name__)


class iris_slack(object):
    supports = frozenset([SLACK_SUPPORT])

    def __init__(self, config):
        self.config = config
        # For now slack has only IM mode but we can expand it to send msg to a
        # channel  instead of a user
        self.modes = {
            SLACK_SUPPORT: self.send_message
        }
        self.proxy = None
        if 'proxy' in self.config:
            host = self.config['proxy']['host']
            port = self.config['proxy']['port']
            self.proxy = {'http': 'http://%s:%s' % (host, port),
                          'https': 'https://%s:%s' % (host, port)}
        self.timeout = config.get('timeout', 10)
        self.sleep_range = config.get('sleep_range', 4)
        self.message_attachments = self.config.get('message_attachments', {})

    def construct_attachments(self, message):
        # TODO: Verify title, title_link and text.
        att_json = {
            'fallback': self.message_attachments.get('fallback'),
            'pretext': self.message_attachments.get('pretext'),
            'title': 'Iris incident %r' % message['incident_id'],
            'mrkdwn_in': ['pretext'],
            'attachment_type': 'default',
            'callback_id': message.get('message_id'),
            'color': 'danger',
            'title_link': '%s/%s' % (
                self.config['iris_incident_url'], message['incident_id']),
            'actions': [
                {
                    'name': 'claim',
                    'text': 'Claim Incident',
                    'type': 'button',
                    'value': 'claimed'
                },
                {
                    'name': 'claim all',
                    'text': 'Claim All',
                    'style': 'danger',
                    'type': 'button',
                    'value': 'claimed all',
                    "confirm": {
                        "title": "Are you sure?",
                        "text": "This will claim all active incidents targeting you.",
                        "ok_text": "Yes",
                        "dismiss_text": "No"
                    }
                }
            ]
        }
        return ujson.dumps([att_json])

    def get_message_payload(self, message):
        slack_message = {
            # Display application if exists, otherwise it's an incident tracking message
            'text': '[%s] %s' % (message.get('application', 'Iris incident'),
                                 message['body']),
            'token': self.config['auth_token'],
            'channel': self.get_destination(message['destination'])
        }
        # only add interactive button for incidents
        if 'incident_id' in message:
            slack_message['attachments'] = self.construct_attachments(message)
        return slack_message

    def get_destination(self, destination):
        # If the destination doesn't have '@' this adds it
        if not (destination.startswith('@') or destination.startswith('#')):
            destination = '@%s' % (destination)
        return destination

    def send_message(self, message):
        start = time.time()
        payload = self.get_message_payload(message)
        try:
            response = requests.post(self.config['base_url'],
                                     data=payload,
                                     proxies=self.proxy,
                                     timeout=self.timeout)
            if response.status_code == 200:
                data = response.json()
                if data['ok']:
                    return time.time() - start
                elif data.get('error') == 'channel_not_found':
                    logger.warning('Slack returned channel_not_found for destination %s', message.get('destination'))
                    return time.time() - start
                # If message is invalid:
                #   {u'ok': False, u'error': u'invalid_arg_name'}
                logger.error('Received an error from slack api: %s',
                             data['error'])
            elif response.status_code == 429:
                # Slack rate limiting. Sleep for a few seconds (chosen randomly to spread load),
                # then raise error to retry
                sleep_time = random.randrange(1, self.sleep_range)
                logger.warning('Hit slack rate limiting, sleeping for %s', sleep_time)
                sleep(sleep_time)
                response.raise_for_status()
            else:
                logger.error('Failed to send message to slack: %d',
                             response.status_code)
        except Exception:
            logger.exception('Slack post request failed')
            raise

    def send(self, message, customizations=None):
        return self.modes[message['mode']](message)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from gevent import monkey, sleep, spawn
monkey.patch_all()  # NOQA

import logging
import logging.handlers
import os
from datetime import datetime

from iris import db, metrics, app_stats
from iris.api import load_config

# logging
logger = logging.getLogger()
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s')
log_file = os.environ.get('APP_STATS_LOG_FILE')
if log_file:
    ch = logging.handlers.RotatingFileHandler(log_file, mode='a', maxBytes=10485760, backupCount=10)
else:
    ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.setLevel(logging.INFO)
logger.addHandler(ch)


# pidfile
pidfile = os.environ.get('APP_STATS_PIDFILE')
if pidfile:
    try:
        pid = os.getpid()
        with open(pidfile, 'w') as h:
            h.write('%s\n' % pid)
            logger.info('Wrote pid %s to %s', pid, pidfile)
    except IOError:
        logger.exception('Failed writing pid to %s', pidfile)

stats_reset = {
    'task_failure': 0,
}


def set_global_stats(stats, connection, cursor):
    # delete outdated stats
    cursor.execute('''DELETE FROM `global_stats`''')

    # format: {statistic : [{timestamp: value}, {timestamp: value}]}
    for stat, entry_list in stats.items():
        for stat_entry in entry_list:
            for timestamp, val in stat_entry.items():
                # format timestamp into datetime to store in mysql
                timestamp = datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
                cursor.execute('''INSERT INTO `global_stats` (`statistic`, `value`, `timestamp`)
                                VALUES (%s, %s, %s)
                                ON DUPLICATE KEY UPDATE `value`= %s, `timestamp` = %s''',
                               (stat, val, timestamp, val, timestamp))
    connection.commit()


def set_app_stats(app, stats, connection, cursor):
    # format: {statistic : [{timestamp: value}, {timestamp: value}]}
    for stat, entry_list in stats.items():
        for stat_entry in entry_list:
            for timestamp, val in stat_entry.items():
                # format timestamp into datetime to store in mysql
                timestamp = datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
                cursor.execute('''INSERT INTO `application_stats` (`application_id`, `statistic`, `value`, `timestamp`)
                                VALUES (%s, %s, %s, %s)
                                ON DUPLICATE KEY UPDATE `value`= %s, `timestamp` = %s''',
                               (app['id'], stat, val, timestamp, val, timestamp))
    connection.commit()


def stats_task(connection, cursor):

    cursor.execute('SELECT `id`, `name` FROM `application`')
    applications = [{'id': row[0], 'name': row[1]} for row in cursor]
    # clean up old app stats
    cursor.execute('''DELETE FROM `application_stats`''')
    for app in applications:
        try:
            stats = app_stats.calculate_app_stats(app, connection, cursor)
            set_app_stats(app, stats, connection, cursor)
        except Exception:
            logger.exception('App stats calculation failed for app %s', app['name'])
            metrics.incr('task_failure')
    try:
        stats = app_stats.calculate_global_stats(connection, cursor)
        set_global_stats(stats, connection, cursor)
    except Exception:
        logger.exception('Global stats calculation failed')
        metrics.incr('task_failure')


def main():
    config = load_config()
    metrics.init(config, 'iris-application-stats', stats_reset)
    app_stats_settings = config.get('app-stats', {})
    run_interval = int(app_stats_settings['run_interval'])
    spawn(metrics.emit_forever)
    db.init(config)
    while True:
        logger.info('Starting app stats calculation loop')
        connection = db.engine.raw_connection()
        cursor = connection.cursor()
        stats_task(connection, cursor)
        cursor.close()
        connection.close()
        logger.info('Waiting %d seconds until next iteration..', run_interval)
        sleep(run_interval)


if __name__ == '__main__':
    main()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from gevent import sleep
from iris.constants import EMAIL_SUPPORT, IM_SUPPORT
from smtplib import SMTP
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
from iris import cache
import quopri
import time
import markdown
import dns.resolver
import dns.exception
import logging

logger = logging.getLogger(__name__)


class iris_smtp(object):
    supports = frozenset([EMAIL_SUPPORT, IM_SUPPORT])

    last_autoscale_mx_lookup = {}

    def __init__(self, config):
        self.config = config
        self.retry_interval = config.get('retry_interval', 0)
        self.modes = {
            EMAIL_SUPPORT: self.send_email,
            IM_SUPPORT: self.send_email,
        }
        self.mx_sorted = []
        self.last_conn = None
        self.last_conn_server = None

        self.smtp_timeout = config.get('timeout', 10)
        if config.get('smtp_server'):
            # mock mx record
            self.mx_sorted.append((0, config['smtp_server']))
        elif config.get('smtp_gateway'):
            try:
                mx_hosts = dns.resolver.query(config['smtp_gateway'], 'MX')
            except dns.exception.DNSException as e:
                mx_hosts = []
                raise Exception('MX error: %s' % e)
            for mx in mx_hosts:
                mx_hostname = mx.exchange.to_text().strip('.')
                self.mx_sorted.append((mx.preference, mx_hostname))
            self.mx_sorted.sort(key=lambda tup: tup[0])
        else:
            raise ValueError('Missing SMTP config for sender')

    def send_email(self, message, customizations=None):
        md = markdown.Markdown()

        start = time.time()
        m = MIMEMultipart('alternative')

        from_address = self.config['from']
        application = message.get('application')
        if application:
            m['X-IRIS-APPLICATION'] = application
            address = cache.applications.get(application, {}).get('custom_sender_addresses', {}).get('email')
            if address is not None:
                from_address = address

        priority = message.get('priority')
        if priority:
            m['X-IRIS-PRIORITY'] = priority

        plan = message.get('plan')
        if plan:
            m['X-IRIS-PLAN'] = plan

        incident_id = message.get('incident_id')
        if incident_id:
            m['X-IRIS-INCIDENT-ID'] = str(incident_id)

        m['Date'] = formatdate(localtime=True)
        m['from'] = from_address
        if message.get('multi-recipient'):
            m['to'] = ','.join(set(message['destination']))
            if message['bcc_destination']:
                m['bcc'] = ','.join(set(message['bcc_destination']))
        else:
            m['to'] = message['destination']
        if message.get('noreply'):
            m['reply-to'] = m['to']

        if 'email_subject' in message:
            m['subject'] = message['email_subject']
        else:
            m['subject'] = message['subject']

        plaintext = None

        if 'email_text' in message:
            plaintext = message['email_text']
        elif 'body' in message:
            plaintext = message['body']

        if plaintext:
            mt = MIMEText(None, 'plain', 'utf-8')
            mt.set_payload(quopri.encodestring(plaintext.encode('UTF-8')))
            mt.replace_header('Content-Transfer-Encoding', 'quoted-printable')
            m.attach(mt)

        # for tracking messages, email_html is not required, so it's possible
        # that both of the following keys are missing from message
        html = None

        if 'email_html' in message:
            html = message['email_html']
        elif 'body' in message:
            html = md.convert(message['body'])

        if html:
            if 'extra_html' in message:
                html += message['extra_html']
            # We need to have body tags for the oneclick buttons to properly parse
            html = '<body>\n' + html + '\n</body>'
            mt = MIMEText(None, 'html', 'utf-8')
            # Google does not like base64 encoded emails for the oneclick button functionalty,
            # so force quoted printable.
            mt.set_payload(quopri.encodestring(html.encode('UTF-8')))
            mt.replace_header('Content-Transfer-Encoding', 'quoted-printable')
            m.attach(mt)

        conn = None

        if message.get('multi-recipient'):
            email_recipients = message['destination'] + message['bcc_destination']
        else:
            email_recipients = [message['destination']]
        # Try reusing previous connection in this worker if we have one
        if self.last_conn:
            conn = self.last_conn
        else:
            for mx in self.mx_sorted:
                try:
                    smtp = SMTP(timeout=self.smtp_timeout)
                    smtp.connect(mx[1], self.config.get('port', 25))
                    if self.config.get('username', None) is not None and self.config.get('password', None) is not None:
                        smtp.login(self.config.get('username', None), self.config.get('password', None))
                    conn = smtp
                    self.last_conn = conn
                    self.last_conn_server = mx[1]
                    break
                except Exception as e:
                    logger.exception(e)

        if not conn:
            raise Exception('Failed to get smtp connection.')

        try:
            conn.sendmail([from_address], email_recipients, m.as_string())
        except Exception:
            logger.warning('Failed sending email through %s. Will try connecting again and resending.', self.last_conn_server)

            try:
                conn.quit()
            except Exception:
                pass

            # If we can't send it, try reconnecting and then sending it one more time before
            # giving up
            for mx in self.mx_sorted:
                try:
                    smtp = SMTP(timeout=self.smtp_timeout)
                    smtp.connect(mx[1], 25)
                    conn = smtp
                    self.last_conn = conn
                    self.last_conn_server = mx[1]
                    break
                except Exception as e:
                    logger.exception('Failed reconnecting to %s to send message', self.last_conn_server)
                    self.last_conn = None
                    return None

            try:
                # If configured, sleep to back-off on connection
                if self.retry_interval:
                    sleep(self.retry_interval)
                conn.sendmail([from_address], email_recipients, m.as_string())
                logger.info('Message successfully sent through %s after reconnecting', self.last_conn_server)
            except Exception:
                logger.exception('Failed sending email through %s after trying to reconnect', self.last_conn_server)
                return None

        return time.time() - start

    def send(self, message, customizations=None):
        return self.modes[message['mode']](message, customizations)

    def cleanup(self):
        if self.last_conn:
            logger.info('Trying to quit smtp connection to %s', self.last_conn_server)
            try:
                self.last_conn.quit()
            except Exception:
                pass

    @classmethod
    def determine_worker_count(cls, vendor):
        mx_gateway = vendor.get('smtp_gateway')
        connections_per_mx = int(vendor.get('tasks_per_mx', 4))

        last_lookup = cls.last_autoscale_mx_lookup.get(mx_gateway)
        now = time.time()

        if last_lookup and last_lookup[0] > now:
            logger.info('Using old MX results for %s; next MX refresh in %d seconds', mx_gateway, last_lookup[0] - now)
            return last_lookup[1]
        else:
            logger.info('Refreshing MX records for %s', mx_gateway)

            try:
                mx_result = dns.resolver.query(mx_gateway, 'MX')
                mx_hosts = [mx.exchange.to_text().strip('.') for mx in mx_result]
            except dns.exception.DNSException as e:
                mx_hosts = []

                if last_lookup:
                    logger.error('Failed looking up MX: %s; returning old results.', e)
                    return last_lookup[1]
                else:
                    raise Exception('MX error, and we don\'t have old results to fall back on: %s' % e)

            mx_host_counts = {mx: connections_per_mx for mx in mx_hosts}
            cls.last_autoscale_mx_lookup[mx_gateway] = (mx_result.expiration, mx_host_counts)
            logger.info('Next MX refresh for %s in %d seconds', mx_gateway, mx_result.expiration - now)
            return mx_host_counts



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from .. import db
from gevent import sleep
import logging
logger = logging.getLogger(__name__)

MODE_CHANGE = 'mode-change'
TARGET_CHANGE = 'target-change'
SENT_CHANGE = 'sent-change'
CONTENT_CHANGE = 'content-change'


def message_change(message_id, change_type, old, new, description):
    if not message_id:
        logger.warning('Not logging %s for message as it does not have an id', change_type)
        return

    old = old[0:250]
    new = new[0:250]
    description = description[0:250]

    # retry to guard against deadlocks
    retries = 0
    max_retries = 5
    while True:
        with db.guarded_session() as session:
            retries += 1
            try:
                session.execute('''
                  INSERT INTO `message_changelog` (`message_id`, `change_type`, `old`, `new`, `description`, `date`)
                  VALUES (:message_id, :change_type, :old, :new, :description, NOW())
                ''', dict(message_id=message_id, change_type=change_type, old=old, new=new, description=description))
                session.commit()
                session.close()
            except Exception:
                logger.exception('Failed inserting message into auditlog. (Try %s/%s)', retries, max_retries)
                if retries < max_retries:
                    sleep(.2)
                    continue
                else:
                    raise Exception('Failed inserting message into auditlog retries exceeded')
            else:
                logger.info('Logged change information for message (ID %s)', message_id)
                break



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import hmac
import hashlib
import urllib.parse
import base64

oneclick_email_markup = '''
<script type="application/ld+json">
{
  "@context": "http://schema.org",
    "@type": "EmailMessage",
    "potentialAction": {
      "@type": "ConfirmAction",
      "name": "Claim",
      "handler": {
        "@type": "HttpActionHandler",
        "url": "%(url)s"
      }
    },
    "description": "Claim incident %(incident_id)s"
}
</script>
'''


def generate_oneclick_url(config, data):
    keys = ('msg_id', 'email_address', 'cmd')  # Order here needs to match order in iris-relay
    HMAC = hmac.new(config['gmail_one_click_url_key'].encode('utf-8'),
                    (' '.join(str(data[key]) for key in keys).encode('utf-8')), hashlib.sha512)
    data['token'] = base64.urlsafe_b64encode(HMAC.digest())
    return config['gmail_one_click_url_endpoint'] + '?' + urllib.parse.urlencode(data)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris.metrics import stats
import logging

logger = logging.getLogger(__name__)

# queue for sending messages. mode -> gevent queue
per_mode_send_queues = {}


def add_mode_stat(mode, runtime):
    try:
        stats[mode + '_cnt'] += 1
        if runtime is None:
            stats[mode + '_fail'] += 1
        else:
            stats[mode + '_total'] += runtime
            stats[mode + '_sent'] += 1
            if runtime > stats[mode + '_max']:
                stats[mode + '_max'] = runtime
            elif runtime < stats[mode + '_min']:
                stats[mode + '_min'] = runtime
            if runtime < stats[mode + '_min']:
                stats[mode + '_min'] = runtime
            elif runtime > stats[mode + '_max']:
                stats[mode + '_max'] = runtime
    except KeyError as e:
        logger.exception('failed modifying nonexistent metric: %s', e)



#!/usr/bin/env python

# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# -*- coding:utf-8 -*-

from contextlib import contextmanager
from sqlalchemy import create_engine
from pymysql.err import IntegrityError
import ujson
import yaml
import click


@click.group()
@click.pass_context
def iris_ctl(ctx):
    pass


@click.group()
@click.pass_context
def app(ctx):
    pass


iris_ctl.add_command(app)


@click.group('import')
@click.pass_context
def app_import(ctx):
    pass


app.add_command(app_import)


def get_db_conn_from_config(config):
    engine = create_engine(config['db']['conn']['str'] % config['db']['conn']['kwargs'],
                           **config['db']['kwargs'])
    return engine.raw_connection()


@contextmanager
def db_from_config(config):
    engine = create_engine(config['db']['conn']['str'] % config['db']['conn']['kwargs'],
                           **config['db']['kwargs'])
    conn = engine.raw_connection()
    cursor = conn.cursor()
    yield conn, cursor
    cursor.close()
    conn.close()


@click.command()
@click.argument('app')
@click.argument('sample_context')
@click.option('--config', default='./config.yaml')
@click.pass_context
def sample_context(ctx, app, sample_context, config):
    with open(sample_context) as fd:
        sample_ctx = fd.read()
        try:
            ujson.loads(sample_ctx)
        except ValueError as e:
            ctx.fail('Invalid JSON, %s: %s' % (str(e), sample_ctx))

    with open(config, 'r') as config_file:
        config = yaml.safe_load(config_file)

    click.echo('Setting sample contenxt for %s to:\n%s' % (app, sample_ctx))
    click.confirm('Do you want to continue?', abort=True)
    with db_from_config(config) as (conn, cursor):
        cursor.execute('UPDATE `application` SET `sample_context`=%s WHERE `name`=%s;',
                       (sample_ctx, app))
        conn.commit()
    click.echo(click.style('All done!', fg='green'))


app_import.add_command(sample_context)


@click.command()
@click.argument('app')
@click.argument('context_template')
@click.option('--config', default='./config.yaml')
@click.pass_context
def context_template(ctx, app, context_template, config):
    with open(context_template) as fd:
        tpl_content = fd.read()

    with open(config, 'r') as config_file:
        config = yaml.safe_load(config_file)

    click.echo('Setting context template for %s to:\n%s' % (app, tpl_content))
    click.confirm('Do you want to continue?', abort=True)
    with db_from_config(config) as (conn, cursor):
        cursor.execute('UPDATE application SET `context_template`=%s WHERE `name`=%s;',
                       (tpl_content, app))
        conn.commit()
    click.echo(click.style('All done!', fg='green'))


app_import.add_command(context_template)


@click.command()
@click.argument('app')
@click.argument('summary_template')
@click.option('--config', default='./config.yaml')
@click.pass_context
def summary_template(ctx, app, summary_template, config):
    with open(summary_template) as fd:
        tpl_content = fd.read()

    with open(config, 'r') as config_file:
        config = yaml.safe_load(config_file)

    click.echo('Setting summary template for %s to:\n%s' % (app, tpl_content))
    click.confirm('Do you want to continue?', abort=True)

    with db_from_config(config) as (conn, cursor):
        cursor.execute('UPDATE application SET `summary_template`=%s WHERE `name`=%s;',
                       (tpl_content, app))
        conn.commit()
    click.echo(click.style('All done!', fg='green'))


app_import.add_command(summary_template)


@click.group()
@click.pass_context
def template(ctx):
    pass


iris_ctl.add_command(template)


@click.command('delete')
@click.argument('template')
@click.option('--config', default='./config.yaml')
@click.pass_context
def delete_template(ctx, template, config):
    with open(config, 'r') as config_file:
        config = yaml.safe_load(config_file)

    click.echo('Deleting template %s' % template)
    click.confirm('Do you want to continue?', abort=True)

    with db_from_config(config) as (conn, cursor):
        try:
            cursor.execute('DELETE FROM template WHERE `name` = %s', template)
            if cursor.rowcount == 0:
                raise click.ClickException('No template found with given name')
        except IntegrityError as e:
            cursor.execute('''SELECT `message`.`id` FROM
                                  message JOIN template ON `message`.`template_id` = `template`.`id`
                              WHERE template.`name` = %s''', template)
            msgs = cursor.fetchall()
            cursor.execute('''SELECT `plan_id` FROM plan_notification JOIN template
                                  ON `plan_notification`.`template_id` = `template`.`id`
                              WHERE template.`name` = %s''', template)
            plans = cursor.fetchall()

            if msgs:
                click.echo(click.style('Template referenced by messages with ids:\n%s' % [m[0] for m in msgs],
                                       fg='red'),
                           err=True)
            if plans:
                click.echo(click.style('Template referenced by plans with ids:\n%s' % [p[0] for p in plans],
                                       fg='red'),
                           err=True)
            raise click.ClickException('Template referenced by a message/plan; for auditing purposes, delete not allowed')
        except Exception as e:
            raise click.ClickException(str(e))
        else:
            conn.commit()
            click.echo(click.style('All done!', fg='green'))


template.add_command(delete_template)


@click.group()
@click.pass_context
def plan(ctx):
    pass


iris_ctl.add_command(plan)


@click.command('delete')
@click.argument('plan')
@click.option('--config', default='./config.yaml')
@click.pass_context
def delete_plan(ctx, plan, config):
    with open(config, 'r') as config_file:
        config = yaml.safe_load(config_file)

    click.echo('Deleting plan %s' % plan)
    click.confirm('Do you want to continue?', abort=True)

    with db_from_config(config) as (conn, cursor):
        try:
            cursor.execute('DELETE plan_active FROM plan_active JOIN plan ON plan_active.`plan_id` = plan.`id` '
                           'WHERE plan.`name` = %s', plan)
            cursor.execute('DELETE plan_notification '
                           'FROM plan_notification JOIN plan ON plan_notification.`plan_id` = plan.`id` '
                           'WHERE plan.`name` = %s', plan)
            cursor.execute('DELETE FROM plan WHERE `name` = %s', plan)
            if cursor.rowcount == 0:
                raise click.ClickException('No plan found with given name')
        except IntegrityError as e:
            cursor.execute('''SELECT `message`.`id` FROM
                                  message JOIN plan ON `message`.`plan_id` = `plan`.`id`
                              WHERE plan.`name` = %s''', plan)
            msgs = cursor.fetchall()
            cursor.execute('''SELECT incident.`id` FROM incident JOIN plan
                                  ON `incident`.`plan_id` = `plan`.`id`
                              WHERE plan.`name` = %s''', plan)
            incidents = cursor.fetchall()

            if msgs:
                click.echo(click.style('Plan referenced by messages with ids:\n%s' % [m[0] for m in msgs],
                                       fg='red'),
                           err=True)
            if incidents:
                click.echo(click.style('Plan referenced by incidents with ids:\n%s' % [i[0] for i in incidents],
                                       fg='red'),
                           err=True)
            raise click.ClickException('Plan referenced by a message/incident; for auditing purposes, delete not allowed')
        except Exception as e:
            raise click.ClickException(str(e))
        else:
            conn.commit()
            click.echo(click.style('All done!', fg='green'))


plan.add_command(delete_plan)


def main():
    iris_ctl(obj={})


if __name__ == '__main__':
    main()



import falcon
from falcon import HTTPNotFound, HTTPFound, HTTPBadRequest
from falcon.util import uri
from jinja2 import FileSystemLoader
from jinja2.sandbox import SandboxedEnvironment
from webassets import Environment as AssetsEnvironment, Bundle
from webassets.ext.jinja2 import AssetsExtension
from webassets.script import CommandLineEnvironment
import os
import ujson
import requests
import importlib
import logging
import re
import pyqrcode
from iris.ui import auth
from beaker.middleware import SessionMiddleware
from datetime import datetime

logger = logging.getLogger(__name__)

_filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')

ui_root = os.environ.get('STATIC_ROOT', os.path.abspath(os.path.dirname(__file__)))


assets_env = AssetsEnvironment(os.path.join(ui_root, 'static'), url='/static')

assets_env.register('jquery_libs', Bundle('js/jquery-3.3.1.min.js', 'js/jquery.dataTables.min.js',
                                          'js/handlebars-4.0.12.min.js', 'js/hopscotch.min.js',
                                          'js/marked.min.js',
                                          output='bundles/jquery.libs.js'))
assets_env.register('bootstrap_libs', Bundle('js/bootstrap.min.js', 'js/typeahead.js',
                                             'js/bootstrap-datetimepicker.js', 'js/moment-timezone.js',
                                             'js/moment-tz-data.js',
                                             output='bundles/bootstrap.libs.js'))
assets_env.register('iris_js', Bundle('js/iris.js', filters='rjsmin', output='bundles/iris.js'))
assets_env.register('css_libs', Bundle('css/bootstrap.min.css', 'css/bootstrap-datetimepicker.css',
                                       'css/jquery.dataTables.min.css', 'css/hopscotch.min.css',
                                       filters='cssmin', output='bundles/libs.css'))

assets_env.register('iris_css', Bundle('css/iris.css', filters='cssmin', output='bundles/iris.css'))

jinja2_env = SandboxedEnvironment(extensions=[AssetsExtension], autoescape=True)
jinja2_env.loader = FileSystemLoader(os.path.join(ui_root, 'templates'))
jinja2_env.assets_environment = assets_env
jinja2_env.filters['tojson'] = ujson.dumps

mimes = {'.css': 'text/css',
         '.jpg': 'image/jpeg',
         '.js': 'text/javascript',
         '.png': 'image/png',
         '.svg': 'image/svg+xml',
         '.ttf': 'application/octet-stream',
         '.woff': 'application/font-woff'}


def hms(seconds):
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return '%dh %02dm %02ds' % (h, m, s)


jinja2_env.filters['hms'] = hms


default_route = '/incidents'


jinja2_env.globals['default_route'] = default_route
jinja2_env.globals['current_year'] = datetime.now().year


def build_assets():
    CommandLineEnvironment(assets_env, logger).build()


def login_url(req):
    if req.path and req.path != '/login' and req.path != '/logout' and req.path != '/':
        return '/login/?next=%s' % uri.encode_value(req.path)
    else:
        return '/login/'


def flash_message(req, message, message_type):
    session = req.env['beaker.session']
    session['flash_message'] = {'type': message_type, 'message': message}
    session.save()


def get_flash(req):
    session = req.env['beaker.session']
    message = session.pop('flash_message', None)
    if message:
        session.save()
        return message
    else:
        return None


# In many cases, these routes need window.appData populated with various bits of data which
# are only retrievable by going through the API itself. Forge a request locally and keep the
# beaker cookies so the current user can be authenticated. This will go away once each page
# that makes use of window.appData is converted to use ajax for those values instead.
def get_local_api(req, path):
    return requests.get('%s/v0/%s' % (local_api_url, path), cookies=req.cookies).json()


def create_qr_code(qr_base_url, qr_login_url):
    qr_code_content = qr_base_url + ',' + qr_login_url
    qr_object = pyqrcode.create(qr_code_content)
    # create qr code and save it as a svg image
    qr_filename = ui_root + '/static/images/iris-mobile-qr.svg'
    qr_object.svg(qr_filename, scale=8)


# Credit to Werkzeug for implementation
def secure_filename(filename):
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')
    filename = str(_filename_ascii_strip_re.sub('', '_'.join(
        filename.split()))).strip('._')
    return filename


class StaticResource(object):
    allow_read_no_auth = True
    frontend_route = False

    def __init__(self, path):
        self.path = path.lstrip('/')

    def on_get(self, req, resp, filename):
        suffix = os.path.splitext(req.path)[1]
        resp.content_type = mimes.get(suffix, 'application/octet-stream')

        filepath = os.path.join(ui_root, self.path, secure_filename(filename))
        try:
            resp.stream = open(filepath, 'rb')
            resp.stream_len = os.path.getsize(filepath)
        except IOError:
            raise HTTPNotFound()


class Index(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        raise HTTPFound(default_route)


class Stats(object):
    allow_read_no_auth = True
    frontend_route = True

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('stats.html').render(request=req)


class AppStats(object):
    allow_read_no_auth = True
    frontend_route = True

    def on_get(self, req, resp, application):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('stats.html').render(request=req)


class SingleStats(object):
    allow_read_no_auth = True
    frontend_route = True

    def on_get(self, req, resp, stat_name):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('singlestat.html').render(request=req)


class Plans(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('plans.html').render(request=req)


class Plan(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp, plan):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('plan.html').render(request=req,
                                                                target_roles=get_local_api(req, 'target_roles'),
                                                                priorities=get_local_api(req, 'priorities'),
                                                                templates=get_local_api(req, 'templates'),
                                                                applications=get_local_api(req, 'applications'))


class Incidents(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('incidents.html').render(request=req,
                                                                     applications=get_local_api(req, 'applications'))


class Incident(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp, incident):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('incident.html').render(request=req)


class Messages(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('messages.html').render(request=req,
                                                                    applications=get_local_api(req, 'applications'),
                                                                    priorities=get_local_api(req, 'priorities'))


class Message(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp, message):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('message.html').render(request=req)


class Templates(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('templates.html').render(request=req)


class Template(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp, template):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('template.html').render(request=req,
                                                                    modes=get_local_api(req, 'modes'),
                                                                    applications=get_local_api(req, 'applications'))


class Applications(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('applications.html').render(request=req,
                                                                        applications=get_local_api(req, 'applications'))


class Application(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp, application):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('application.html').render(request=req,
                                                                       applications=get_local_api(req, 'applications'),
                                                                       priorities=get_local_api(req, 'priorities'),
                                                                       modes=get_local_api(req, 'modes') + ['drop'])


class Unsubscribe(object):
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp, application):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('unsubscribe.html').render(request=req,
                                                                       priorities=get_local_api(req, 'priorities'))


class Login():
    allow_read_no_auth = False
    frontend_route = True

    def __init__(self, auth_manager, debug):
        self.auth_manager = auth_manager
        self.debug = debug

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('login.html').render(request=req,
                                                                 path='login',
                                                                 last_flash=get_flash(req))

    def on_post(self, req, resp):
        form_body = uri.parse_query_string(req.context['body'].decode('utf-8'))

        try:
            username = form_body['username']
            password = form_body['password']
        except KeyError:
            raise HTTPFound('/login')

        if not auth.valid_username(username):
            logger.warning('Tried to login with invalid username %s', username)
            if self.debug:
                flash_message(req, 'Invalid username', 'danger')
            else:
                flash_message(req, 'Invalid credentials', 'danger')
            raise HTTPFound('/login')

        if self.auth_manager.authenticate(username, password):
            logger.info('Successful login for %s', username)
            auth.login_user(req, username)
        else:
            logger.warning('Failed login for %s', username)
            flash_message(req, 'Invalid credentials', 'danger')
            raise HTTPFound('/login')

        # Remove newlines to prevent HTTP request splitting
        url = req.get_param('next', default='').replace('\n', '')

        if not url or url.startswith('/'):
            raise HTTPFound(url or default_route)
        else:
            raise HTTPBadRequest('Invalid next parameter', '')


class Logout():
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        auth.logout_user(req)
        raise HTTPFound('/login')


class User():
    allow_read_no_auth = False
    frontend_route = True

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('user.html').render(request=req,
                                                                modes=get_local_api(req, 'modes'),
                                                                priorities=get_local_api(req, 'priorities'),
                                                                applications=get_local_api(req, 'applications'))


class Qr(object):
    allow_read_no_auth = True
    frontend_route = True

    def __init__(self, qr_base_url, qr_login_url):
        self.qr_base_url = qr_base_url
        self.qr_login_url = qr_login_url

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = jinja2_env.get_template('qr.html').render(base=self.qr_base_url, login=self.qr_login_url)


class JinjaValidate():
    allow_read_no_auth = False
    frontend_route = True

    def on_post(self, req, resp):
        form_body = uri.parse_query_string(req.context['body'].decode('utf-8'))

        try:
            template_subject = form_body['templateSubject']
            template_body = form_body['templateBody']
            application = form_body['application']
        except KeyError:
            raise HTTPBadRequest('Missing keys from post body', '')

        if not application:
            resp.body = ujson.dumps({'error': 'No application found'})
            resp.status = falcon.HTTP_400
            return

        app_json = get_local_api(req, 'applications/%s' % application)
        sample_context_str = app_json.get('sample_context')
        if not sample_context_str:
            resp.body = ujson.dumps({'error': 'Missing sample_context from application config'})
            resp.status = falcon.HTTP_400
            logger.error('Missing sample context for app %s', application)
            return

        try:
            sample_context = ujson.loads(sample_context_str)
        except Exception:
            resp.body = ujson.dumps({'error': 'Invalid application sample_context'})
            resp.status = falcon.HTTP_400
            logger.exception('Bad sample context for app %s', application)
            return

        # TODO: also move iris meta var to api
        iris_sample_context = {
            "message_id": 5456900,
            "target": "user",
            "priority": "Urgent",
            "application": "Autoalerts",
            "plan": "default plan",
            "plan_id": 1843,
            "incident_id": 178293332,
            "template": "default template"
        }
        sample_context['iris'] = iris_sample_context

        environment = SandboxedEnvironment()

        try:
            subject_template = environment.from_string(template_subject)
            body_template = environment.from_string(template_body)
        except Exception as e:
            resp.body = ujson.dumps({'error': str(e), 'lineno': e.lineno})
            resp.status = falcon.HTTP_400
            return

        try:
            rendered_subject = subject_template.render(sample_context),
            rendered_body = body_template.render(sample_context)
        except Exception as e:
            resp.body = ujson.dumps({'error': str(e)})
            resp.status = falcon.HTTP_400
            return

        resp.body = ujson.dumps({
            'template_subject': rendered_subject,
            'template_body': rendered_body
        })


def init(config, app):
    global local_api_url
    logger.info('Web asset root: "%s"', ui_root)
    auth_module = config.get('auth', {'module': 'iris.ui.auth.noauth'})['module']
    auth = importlib.import_module(auth_module)
    auth_manager = getattr(auth, 'Authenticator')(config)
    qr_base_url = config.get('qr_base_url')
    qr_login_url = config.get('qr_login_url')

    debug = config['server'].get('disable_auth', False) is True
    local_api_url = config['server'].get('local_api_url', 'http://localhost:16649')

    app.add_route('/static/bundles/{filename}', StaticResource('/static/bundles'))
    app.add_route('/static/images/{filename}', StaticResource('/static/images'))
    app.add_route('/static/fonts/{filename}', StaticResource('/static/fonts'))
    app.add_route('/', Index())
    app.add_route('/stats', Stats())
    app.add_route('/stats/{application}', AppStats())
    app.add_route('/singlestats/{stat_name}', SingleStats())
    app.add_route('/plans/', Plans())
    app.add_route('/plans/{plan}', Plan())
    app.add_route('/incidents/', Incidents())
    app.add_route('/incidents/{incident}', Incident())
    app.add_route('/messages/', Messages())
    app.add_route('/messages/{message}', Message())
    app.add_route('/templates/', Templates())
    app.add_route('/templates/{template}', Template())
    app.add_route('/applications/', Applications())
    app.add_route('/applications/{application}', Application())
    app.add_route('/login/', Login(auth_manager, debug))
    app.add_route('/logout/', Logout())
    app.add_route('/user/', User())
    app.add_route('/validate/jinja', JinjaValidate())
    app.add_route('/unsubscribe/{application}', Unsubscribe())

    if(qr_base_url and qr_login_url):
        create_qr_code(qr_base_url, qr_login_url)
        app.add_route('/qr', Qr(qr_base_url, qr_login_url))

    # Configuring the beaker middleware mutilates the app object, so do it
    # at the end, after we've added all routes/sinks for the entire iris
    # app.
    session_opts = {
        'session.type': 'cookie',
        'session.cookie_expires': True,
        'session.key': 'iris-auth',
        'session.encrypt_key': config['user_session']['encrypt_key'],
        'session.validate_key': config['user_session']['sign_key'],
        'session.secure': not (config['server'].get('disable_auth', False) or config['server'].get('allow_http', False)),
        'session.httponly': True,
        'session.crypto_type': 'cryptography',
        'session.samesite': 'Lax'
    }
    app = SessionMiddleware(app, session_opts)

    return app



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import os
from gevent import Timeout, socket
from gevent.server import StreamServer
import msgpack
from ..utils import msgpack_unpack_msg_from_socket, sanitize_unicode_dict
from . import cache
from iris import metrics
from iris.role_lookup import IrisRoleLookupException

import logging
import logging.handlers
logger = logging.getLogger(__name__)
access_logger = logging.getLogger('RPC:access')


send_funcs = {}
rpc_timeout = None
rpc_server = None


def msgpack_handle_sets(obj):
    if isinstance(obj, set):
        return list(obj)


def generate_msgpack_message_payload(message):
    return msgpack.packb({'endpoint': 'v0/slave_send', 'data': message}, default=msgpack_handle_sets)


def send_message_to_slave(message, address):
    try:
        payload = generate_msgpack_message_payload(message)
    except TypeError:
        logger.exception('Failed encoding message %s as msgpack', message)
        metrics.incr('rpc_message_pass_fail_cnt')
        return False

    pretty_address = '%s:%s' % address
    message_id = message.get('message_id', '?')
    try:
        s = socket.create_connection(address)
        s.send(payload)
        sender_resp = msgpack_unpack_msg_from_socket(s)
        s.close()
    except socket.error:
        logging.exception('Failed connecting to %s to send message (ID %s)',
                          pretty_address, message_id)
        metrics.incr('rpc_message_pass_fail_cnt')
        return False

    if sender_resp == 'OK':
        access_logger.info('Successfully passed message (ID %s) to %s for sending',
                           message_id, pretty_address)
        metrics.incr('rpc_message_pass_success_cnt')
        return True
    else:
        logger.error('Failed sending message (ID %s) through %s: %s',
                     message_id, pretty_address, sender_resp)
        metrics.incr('rpc_message_pass_fail_cnt')
        return False


def reject_api_request(socket, address, err_msg):
    logger.info('-> %s %s', address, err_msg)
    socket.sendall(msgpack.packb(err_msg))


def handle_api_notification_request(socket, address, req):
    notification = req['data']
    if 'application' not in notification:
        reject_api_request(socket, address, 'INVALID application')
        logger.warning('Dropping OOB message due to missing application key')
        return
    notification['subject'] = '[%s] %s' % (notification['application'],
                                           notification.get('subject', ''))
    target_list = notification.get('target_list')
    role = notification.get('role')
    if not role and not target_list:
        reject_api_request(socket, address, 'INVALID role')
        logger.warning('Dropping OOB message with invalid role "%s" from app %s',
                       role, notification['application'])
        return
    target = notification.get('target')
    if not (target or target_list):
        reject_api_request(socket, address, 'INVALID target')
        logger.warning('Dropping OOB message with invalid target "%s" from app %s',
                       target, notification['application'])
        return
    expanded_targets = None
    # if role is literal_target skip unrolling
    if not notification.get('unexpanded'):
        # For multi-recipient notifications, pre-populate destination with literal targets,
        # then expand the remaining
        has_literal_target = False
        if target_list:
            expanded_targets = []
            notification['destination'] = []
            notification['bcc_destination'] = []
            for t in target_list:
                role = t['role']
                target = t['target']
                bcc = t.get('bcc')
                try:
                    if role == 'literal_target':
                        if bcc:
                            notification['bcc_destination'].append(target)
                        else:
                            notification['destination'].append(target)
                        has_literal_target = True
                    else:
                        expanded = cache.targets_for_role(role, target)
                        expanded_targets += [{'target': e, 'bcc': bcc} for e in expanded]
                except IrisRoleLookupException:
                    # Maintain best-effort delivery for remaining targets if one fails to resolve
                    continue
        else:
            try:
                expanded_targets = cache.targets_for_role(role, target)
            except IrisRoleLookupException:
                expanded_targets = None
        if not expanded_targets and not has_literal_target:
            reject_api_request(socket, address, 'INVALID role:target')
            logger.warning('Dropping OOB message with invalid role:target "%s:%s" from app %s',
                           role, target, notification['application'])
            return

    sanitize_unicode_dict(notification)

    # If we're rendering this using templates+context instead of body, fill in the
    # needed iris key.
    if 'template' in notification:
        if 'context' not in notification:
            logger.warning('Dropping OOB message due to missing context from app %s',
                           notification['application'])
            reject_api_request(socket, address, 'INVALID context')
            return
        else:
            # fill in dummy iris meta data
            notification['context']['iris'] = {}
    elif 'email_html' in notification:
        if not isinstance(notification['email_html'], str):
            logger.warning('Dropping OOB message with invalid email_html from app %s: %s',
                           notification['application'], notification['email_html'])
            reject_api_request(socket, address, 'INVALID email_html')
            return
    elif 'body' not in notification:
        reject_api_request(socket, address, 'INVALID body')
        logger.warning('Dropping OOB message with invalid body from app %s',
                       notification['application'])
        return

    access_logger.info('-> %s OK, to %s:%s (%s:%s)',
                       address, role, target, notification['application'],
                       notification.get('priority', notification.get('mode', '?')))

    notification_count = 1
    if notification.get('unexpanded'):
        notification['destination'] = notification['target']
        send_funcs['message_send_enqueue'](notification)
    elif notification.get('multi-recipient'):
        notification['target'] = expanded_targets
        send_funcs['message_send_enqueue'](notification)
        notification_count = len(expanded_targets)
    else:
        for _target in expanded_targets:
            temp_notification = notification.copy()
            temp_notification['target'] = _target
            send_funcs['message_send_enqueue'](temp_notification)
    metrics.incr('notification_cnt', inc=notification_count)
    socket.sendall(msgpack.packb('OK'))


def handle_slave_send(socket, address, req):
    message = req['data']
    message_id = message.get('message_id', '?')

    message['to_slave'] = True

    try:
        runtime = send_funcs['message_send_enqueue'](message)
        response = 'OK'
        access_logger.info('Message (ID %s) from master %s queued successfully', message_id, address)
    except Exception:
        response = 'FAIL'
        logger.exception('Queueing message (ID %s) from master %s failed.')
        access_logger.error('Failed queueing message (ID %s) from master %s: %s', message_id, address, runtime)
        metrics.incr('slave_message_send_fail_cnt')

    socket.sendall(msgpack.packb(response))


api_request_handlers = {
    'v0/send': handle_api_notification_request,
    'v0/slave_send': handle_slave_send
}


def handle_api_request(socket, address):
    metrics.incr('api_request_cnt')
    timeout = Timeout.start_new(rpc_timeout)
    try:
        req = msgpack_unpack_msg_from_socket(socket)
        if not req:
            logger.warning('Couldn\'t get msgpack data from %s', address)
            socket.close()
            return
        access_logger.info('%s %s', address, req['endpoint'])
        handler = api_request_handlers.get(req['endpoint'])
        if handler is not None:
            handler(socket, address, req)
        else:
            logger.info('-> %s unknown request', address)
            socket.sendall(msgpack.packb('UNKNOWN'))
    except Timeout:
        metrics.incr('api_request_timeout_cnt')
        logger.warning('-> %s timeout', address)
        socket.sendall(msgpack.packb('TIMEOUT'))
    finally:
        timeout.cancel()
    socket.close()


def run(sender_config):
    global rpc_server
    try:
        rpc_server = StreamServer((sender_config['host'], sender_config['port']),
                                  handle_api_request)
        rpc_server.start()
        return True
    except Exception:
        logger.exception('Failed binding to sender RPC port')
        return False


def shutdown():
    global rpc_server
    if rpc_server:
        logger.info('Stopping RPC server')
        rpc_server.close()


def init(sender_config, _send_funcs):
    global rpc_timeout

    send_funcs.update(_send_funcs)

    default_rpc_timeout = 20
    try:
        rpc_timeout = int(sender_config.get('rpc_timeout', default_rpc_timeout))
    except ValueError:
        logger.exception('Failed parsing rpc_timeout in config')
        rpc_timeout = default_rpc_timeout
    logger.info('RPC timeout is set to %s seconds', rpc_timeout)

    access_log_cfg = {
        'filename': './access.log',
        'mode': 'a',
        'maxBytes': 104857600,
        'backupCount': 20,
    }
    access_log_cfg.update(sender_config.get('access_log', {}))

    log_dir = os.path.dirname(access_log_cfg['filename'])
    if not os.path.exists(log_dir):
        os.mkdir(log_dir)

    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    log_hdl = logging.handlers.RotatingFileHandler(**access_log_cfg)
    log_hdl.setFormatter(formatter)
    access_logger.propagate = False
    access_logger.addHandler(log_hdl)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from gevent import monkey, sleep, spawn
monkey.patch_all()  # NOQA

import logging
import logging.handlers
import os
import uuid
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import SQLAlchemyError, IntegrityError, DataError
import requests
import oncallclient
from phonenumbers import format_number, parse, PhoneNumberFormat
from phonenumbers.phonenumberutil import NumberParseException

from iris.api import load_config
from iris import metrics

# Used for the optional ldap mailing list resolving functionality
import ldap
from ldap.controls import SimplePagedResultsControl
from ldap.filter import escape_filter_chars
import time
ldap_pagination_size = None
ldap_timeout = None
update_sleep = None

logging.getLogger('requests').setLevel(logging.WARNING)


stats_reset = {
    'sql_errors': 0,
    'users_added': 0,
    'users_failed_to_add': 0,
    'users_failed_to_update': 0,
    'users_purged': 0,
    'others_purged': 0,
    'teams_added': 0,
    'teams_failed_to_add': 0,
    'user_contacts_updated': 0,
    'ldap_lists_added': 0,
    'ldap_memberships_added': 0,
    'ldap_lists_removed': 0,
    'ldap_memberships_removed': 0,
    'ldap_lists_failed_to_add': 0,
    'ldap_memberships_failed_to_add': 0,
    'ldap_reconnects': 0
}

# logging
logger = logging.getLogger()
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s')
log_file = os.environ.get('SYNC_TARGETS_LOG_FILE')
if log_file:
    ch = logging.handlers.RotatingFileHandler(log_file, mode='a', maxBytes=10485760, backupCount=10)
else:
    ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.setLevel(logging.INFO)
logger.addHandler(ch)

pidfile = os.environ.get('SYNC_TARGETS_PIDFILE')
if pidfile:
    try:
        pid = os.getpid()
        with open(pidfile, 'w') as h:
            h.write('%s\n' % pid)
            logger.info('Wrote pid %s to %s', pid, pidfile)
    except IOError:
        logger.exception('Failed writing pid to %s', pidfile)


def normalize_phone_number(num):
    return format_number(parse(num, 'US'),
                         PhoneNumberFormat.INTERNATIONAL)


def get_predefined_users(config):
    users = {}
    config_users = []

    try:
        config_users = config['sync_script']['preset_users']
    except KeyError:
        return {}

    for user in config_users:
        users[user['name']] = user
        for key in ('sms', 'call'):
            try:
                users[user['name']][key] = normalize_phone_number(users[user['name']][key])
            except (NumberParseException, KeyError, AttributeError):
                users[user['name']][key] = None

    return users


def prune_target(engine, target_name, target_type):
    if target_type == 'user':
        metrics.incr('users_purged')
    else:
        metrics.incr('others_purged')

    if target_type == 'team':
        try:
            # rename team to prevent namespace conflict but preserve the ability to reactivate it in the future
            new_name = str(uuid.uuid4())
            engine.execute('''UPDATE `target` SET `active` = FALSE, `name` = %s WHERE `name` = %s AND `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = %s)''', (new_name, target_name, target_type))
            logger.info('Deleted inactive oncall team %s', target_name)
        except SQLAlchemyError as e:
            logger.error('Deleting oncall team %s failed: %s', target_name, e)
            metrics.incr('sql_errors')
        return

    try:
        engine.execute('''DELETE FROM `target` WHERE `name` = %s AND `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = %s)''', (target_name, target_type))
        logger.info('Deleted inactive target %s', target_name)

    # The user has messages or some other user data which should be preserved.
    # Just mark as inactive.
    except IntegrityError:
        logger.info('Marking target %s inactive', target_name)
        engine.execute('''UPDATE `target` SET `active` = FALSE WHERE `name` = %s AND `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = %s)''', (target_name, target_type))

    except SQLAlchemyError as e:
        logger.error('Deleting target %s failed: %s', target_name, e)
        metrics.incr('sql_errors')
        return


def fetch_teams_from_oncall(oncall):
    try:
        return oncall.get('%steams?fields=name&active=1&get_id=1' % oncall.url).json()
    except (ValueError, requests.exceptions.RequestException):
        logger.exception('Failed hitting oncall endpoint to fetch list of team names')
        return []


def fix_user_contacts(contacts):
    sms = contacts.get('sms')
    if sms:
        contacts['sms'] = normalize_phone_number(sms)

    call = contacts.get('call')
    if call:
        contacts['call'] = normalize_phone_number(call)

    return contacts


def fetch_users_from_oncall(oncall):
    oncall_user_endpoint = oncall.url + 'users?fields=name&fields=contacts&fields=active'
    try:
        return {user['name']: fix_user_contacts(user['contacts'])
                for user in oncall.get(oncall_user_endpoint).json()
                if user['active']}
    except (ValueError, KeyError, requests.exceptions.RequestException):
        logger.exception('Failed hitting oncall endpoint to fetch list of users')
        return {}


def sync_from_oncall(config, engine, purge_old_users=True):
    # users and teams present in our oncall database
    oncall_base_url = config.get('oncall-api')

    if not oncall_base_url:
        logger.error('Missing URL to oncall-api, which we use for user/team lookups. Bailing.')
        return

    oncall = oncallclient.OncallClient(config.get('oncall-app', ''), config.get('oncall-key', ''), oncall_base_url)
    oncall_users = fetch_users_from_oncall(oncall)

    if not oncall_users:
        logger.warning('No users found. Bailing.')
        return

    # get teams from oncall-api and separate the list of tuples into two lists of name and ids
    oncall_teams_api_response = fetch_teams_from_oncall(oncall)
    if not oncall_teams_api_response:
        logger.warning('No teams found. Bailing.')
        return

    oncall_team_response = list(zip(*oncall_teams_api_response))
    oncall_team_names = [name.lower() for name in oncall_team_response[0]]
    oncall_team_ids = oncall_team_response[1]
    oncall_response_dict_name_key = dict(zip(oncall_team_names, oncall_team_ids))
    oncall_response_dict_id_key = dict(zip(oncall_team_ids, oncall_team_names))
    oncall_case_sensitive_dict = {name.lower(): name for name in oncall_team_response[0]}

    if not oncall_team_names:
        logger.warning('We do not have a list of team names')

    oncall_team_names = set(oncall_team_names)
    oncall_team_ids = set(oncall_team_ids)

    session = sessionmaker(bind=engine)()

    # users present in iris' database
    iris_users = {}
    for row in engine.execute('''SELECT `target`.`name` as `name`, `mode`.`name` as `mode`,
                                        `target_contact`.`destination`
                                 FROM `target`
                                 JOIN `user` on `target`.`id` = `user`.`target_id`
                                 LEFT OUTER JOIN `target_contact` ON `target`.`id` = `target_contact`.`target_id`
                                 LEFT OUTER JOIN `mode` ON `target_contact`.`mode_id` = `mode`.`id`
                                 WHERE `target`.`active` = TRUE
                                 ORDER BY `target`.`name`'''):
        contacts = iris_users.setdefault(row.name, {})
        if row.mode is None or row.destination is None:
            continue
        contacts[row.mode] = row.destination

    iris_usernames = iris_users.keys()

    # users from the oncall endpoints and config files
    metrics.set('users_found', len(oncall_users))
    metrics.set('teams_found', len(oncall_team_names))
    oncall_users.update(get_predefined_users(config))
    oncall_usernames = oncall_users.keys()

    # set of users not presently in iris
    users_to_insert = oncall_usernames - iris_usernames
    # set of existing iris users that are in the user oncall database
    users_to_update = iris_usernames & oncall_usernames
    users_to_mark_inactive = iris_usernames - oncall_usernames

    # get objects needed for insertion
    target_types = {name: target_id for name, target_id in session.execute('SELECT `name`, `id` FROM `target_type`')}  # 'team' and 'user'
    modes = {name: mode_id for name, mode_id in session.execute('SELECT `name`, `id` FROM `mode`')}
    iris_team_names = {name.lower() for (name, ) in engine.execute('''SELECT `name` FROM `target` WHERE `type_id` = %s''', target_types['team'])}
    target_add_sql = 'INSERT INTO `target` (`name`, `type_id`) VALUES (%s, %s) ON DUPLICATE KEY UPDATE `active` = TRUE'
    oncall_add_sql = 'INSERT INTO `oncall_team` (`target_id`, `oncall_team_id`) VALUES (%s, %s)'
    user_add_sql = 'INSERT IGNORE INTO `user` (`target_id`) VALUES (%s)'
    target_contact_add_sql = '''INSERT INTO `target_contact` (`target_id`, `mode_id`, `destination`)
                                VALUES (%s, %s, %s)
                                ON DUPLICATE KEY UPDATE `destination` = %s'''

    # insert users that need to be
    logger.info('Users to insert (%d)', len(users_to_insert))
    for username in users_to_insert:
        sleep(update_sleep)
        logger.info('Inserting %s', username)
        try:
            target_id = engine.execute(target_add_sql, (username, target_types['user'])).lastrowid
            engine.execute(user_add_sql, (target_id, ))
        except SQLAlchemyError as e:
            metrics.incr('users_failed_to_add')
            metrics.incr('sql_errors')
            logger.exception('Failed to add user %s' % username)
            continue
        metrics.incr('users_added')
        for key, value in oncall_users[username].items():
            if value and key in modes:
                logger.info('%s: %s -> %s', username, key, value)
                engine.execute(target_contact_add_sql, (target_id, modes[key], value, value))

    # update users that need to be
    contact_update_sql = 'UPDATE target_contact SET destination = %s WHERE target_id = (SELECT id FROM target WHERE name = %s AND type_id = %s) AND mode_id = %s'
    contact_insert_sql = 'INSERT INTO target_contact (target_id, mode_id, destination) VALUES ((SELECT id FROM target WHERE name = %s AND type_id = %s), %s, %s)'
    contact_delete_sql = 'DELETE FROM target_contact WHERE target_id = (SELECT id FROM target WHERE name = %s AND type_id = %s) AND mode_id = %s'

    logger.info('Users to update (%d)', len(users_to_update))
    for username in users_to_update:
        sleep(update_sleep)
        try:
            db_contacts = iris_users[username]
            oncall_contacts = oncall_users[username]
            for mode in modes:
                if mode in oncall_contacts and oncall_contacts[mode]:
                    if mode in db_contacts:
                        if oncall_contacts[mode] != db_contacts[mode]:
                            logger.info('%s: updating %s', username, mode)
                            metrics.incr('user_contacts_updated')
                            engine.execute(contact_update_sql, (oncall_contacts[mode], username, target_types['user'], modes[mode]))
                    else:
                        logger.info('%s: adding %s', username, mode)
                        metrics.incr('user_contacts_updated')
                        engine.execute(contact_insert_sql, (username, target_types['user'], modes[mode], oncall_contacts[mode]))
                elif mode in db_contacts:
                    logger.info('%s: deleting %s', username, mode)
                    metrics.incr('user_contacts_updated')
                    engine.execute(contact_delete_sql, (username, target_types['user'], modes[mode]))
                else:
                    logger.debug('%s: missing %s', username, mode)
        except SQLAlchemyError as e:
            metrics.incr('users_failed_to_update')
            metrics.incr('sql_errors')
            logger.exception('Failed to update user %s', username)
            continue

# sync teams between iris and oncall

    # iris_db_oncall_team_ids (team_ids in the oncall_team table)
    # oncall_team_ids (team_ids from oncall api call)
    # oncall_team_names (names from oncall api call)
    # oncall_response_dict_name_key (key value pairs of oncall team names and ids from api call)
    # oncall_response_dict_id_key same as above but key value inverted
    # oncall_case_sensitive_dict maps the case insensitive oncall name to the original capitalization
    # iris_team_names (names from target table)
    # iris_target_name_id_dict dictionary of target name -> target_id mappings
    # iris_db_oncall_team_id_name_dict dictionary of oncall team_id -> oncall name mappings

# get all incoming names that match a target check if that target has an entry in oncall table if not make one
    iris_target_name_id_dict = {name.lower(): target_id for name, target_id in engine.execute('''SELECT `name`, `id` FROM `target` WHERE `type_id` = %s''', target_types['team'])}

    matching_target_names = iris_team_names.intersection(oncall_team_names)
    if matching_target_names:
        existing_up_to_date_oncall_teams = {name.lower() for (name, ) in session.execute('''SELECT `target`.`name` FROM `target` JOIN `oncall_team` ON `oncall_team`.`target_id` = `target`.`id` WHERE `target`.`name` IN :matching_names''', {'matching_names': tuple(matching_target_names)})}
        # up to date target names that don't have an entry in the oncall_team table yet
        matching_target_names_no_oncall_entry = matching_target_names - existing_up_to_date_oncall_teams

        for t in matching_target_names_no_oncall_entry:
            logger.info('Inserting existing team into oncall_team %s', t)
            sleep(update_sleep)
            try:
                engine.execute('''UPDATE `target` SET `active` = TRUE WHERE `id` = %s''', iris_target_name_id_dict[t])
                engine.execute(oncall_add_sql, (iris_target_name_id_dict[t], oncall_response_dict_name_key[t]))
            except SQLAlchemyError as e:
                logger.exception('Error inserting oncall_team %s: %s', t, e)
                continue

# rename all mismatching target names

    iris_db_oncall_team_id_name_dict = {team_id: name.lower() for name, team_id in engine.execute('''SELECT target.name, oncall_team.oncall_team_id FROM `target` JOIN `oncall_team` ON oncall_team.target_id = target.id''')}

    iris_db_oncall_team_ids = {oncall_team_id for (oncall_team_id, ) in engine.execute('''SELECT `oncall_team_id` FROM `oncall_team`''')}
    matching_oncall_ids = oncall_team_ids.intersection(iris_db_oncall_team_ids)

    name_swaps = {}

    # find teams in the iris database whose names have changed
    for oncall_id in matching_oncall_ids:

        current_name = iris_db_oncall_team_id_name_dict[oncall_id]
        new_name = oncall_response_dict_id_key[oncall_id]
        try:
            if current_name != new_name:
                # handle edge case of teams swapping names
                if not iris_target_name_id_dict.get(new_name, None):
                    target_id_to_rename = iris_target_name_id_dict[current_name]
                    logger.info('Renaming team %s to %s', current_name, new_name)
                    engine.execute('''UPDATE `target` SET `name` = %s, `active` = TRUE WHERE `id` = %s''', (oncall_case_sensitive_dict[new_name], target_id_to_rename))
                else:
                    # there is a team swap so rename to a random name to prevent a violation of unique target name constraint
                    new_name = str(uuid.uuid4())
                    target_id_to_rename = iris_target_name_id_dict[current_name]
                    name_swaps[oncall_id] = target_id_to_rename
                    logger.info('Renaming team %s to %s', current_name, new_name)
                    engine.execute('''UPDATE `target` SET `name` = %s, `active` = TRUE WHERE `id` = %s''', (new_name, target_id_to_rename))
                sleep(update_sleep)
        except Exception:
            logger.exception('Error changing team name of %s to %s', current_name, new_name)

    # go back and rename name_swaps to correct value
    for oncall_id, target_id_to_rename in name_swaps.items():
        new_name = oncall_response_dict_id_key[oncall_id]
        engine.execute('''UPDATE `target` SET `name` = %s, `active` = TRUE WHERE `id` = %s''', (oncall_case_sensitive_dict[new_name], target_id_to_rename))
        sleep(update_sleep)


# create new entries for new teams

    # if the team_id doesn't exist in oncall_team at this point then it is a new team.
    new_team_ids = oncall_team_ids - iris_db_oncall_team_ids
    logger.info('Teams to insert (%d)' % len(new_team_ids))

    for team_id in new_team_ids:
        t = oncall_case_sensitive_dict[oncall_response_dict_id_key[team_id]]
        new_target_id = None

        # add team to target table
        logger.info('Inserting team %s', t)
        sleep(update_sleep)
        try:
            new_target_id = engine.execute(target_add_sql, (t, target_types['team'])).lastrowid
            metrics.incr('teams_added')
        except SQLAlchemyError as e:
            logger.exception('Error inserting team %s: %s', t, e)
            metrics.incr('teams_failed_to_add')
            continue

        # add team to oncall_team table
        if new_target_id:
            logger.info('Inserting new team into oncall_team %s', t)
            try:
                engine.execute(oncall_add_sql, (new_target_id, team_id))
            except SQLAlchemyError as e:
                logger.exception('Error inserting oncall_team %s: %s', t, e)
                continue
    session.commit()
    session.close()

    # mark users/teams inactive
    if purge_old_users:
        # find active teams that don't exist in oncall anymore
        updated_iris_team_names = {name.lower() for (name, ) in engine.execute('''SELECT `name` FROM `target` WHERE `type_id` = %s AND `active` = TRUE''', target_types['team'])}
        teams_to_deactivate = updated_iris_team_names - oncall_team_names

        logger.info('Users to mark inactive (%d)' % len(users_to_mark_inactive))
        logger.info('Teams to mark inactive (%d)' % len(teams_to_deactivate))
        for username in users_to_mark_inactive:
            prune_target(engine, username, 'user')
            sleep(update_sleep)
        for team in teams_to_deactivate:
            prune_target(engine, team, 'team')
            sleep(update_sleep)


def get_ldap_lists(l, search_strings, parent_list=None):
    results = set()
    known_ldap_resp_ctrls = {
        SimplePagedResultsControl.controlType: SimplePagedResultsControl,
    }
    req_ctrl = SimplePagedResultsControl(True, size=ldap_pagination_size, cookie='')

    if parent_list:
        filterstr = search_strings['get_all_sub_lists_filter'] % escape_filter_chars(parent_list)
    else:
        filterstr = search_strings['get_all_lists_filter']

    while True:
        msgid = l.search_ext(search_strings['list_search_string'],
                             ldap.SCOPE_SUBTREE,
                             serverctrls=[req_ctrl],
                             attrlist=(search_strings['list_cn_field'], search_strings['list_name_field']),
                             filterstr=filterstr)

        retries = 0
        rdata = None
        while retries < 5:
            try:
                rtype, rdata, rmsgid, serverctrls = l.result3(msgid, timeout=ldap_timeout, resp_ctrl_classes=known_ldap_resp_ctrls)
            except Exception:
                logger.exception('Error getting ldap list memberships')
                retries += 1
            else:
                break
        # ldap sometimees returns good but does not return rdata
        if rdata is None:
            break
        for (dn, data) in rdata:
            cn_field = data[search_strings['list_cn_field']][0]
            name_field = data[search_strings['list_name_field']][0]

            if isinstance(cn_field, bytes):
                cn_field = cn_field.decode('utf-8')
            if isinstance(name_field, bytes):
                name_field = name_field.decode('utf-8')

            results |= {(cn_field, name_field)}

        pctrls = [c for c in serverctrls
                  if c.controlType == SimplePagedResultsControl.controlType]

        if not pctrls:
            # Paging not supported
            logger.debug('LDAP pagination not supported')
            break
        cookie = pctrls[0].cookie
        if not cookie:
            break
        req_ctrl.cookie = cookie

    return results


def get_ldap_list_membership(l, search_strings, list_name):
    results = set()
    known_ldap_resp_ctrls = {
        SimplePagedResultsControl.controlType: SimplePagedResultsControl,
    }
    req_ctrl = SimplePagedResultsControl(True, size=ldap_pagination_size, cookie='')
    while True:
        msgid = l.search_ext(search_strings['user_search_string'],
                             ldap.SCOPE_SUBTREE,
                             serverctrls=[req_ctrl],
                             attrlist=[search_strings['user_mail_field']],
                             filterstr=search_strings['user_membership_filter'] % escape_filter_chars(list_name)
                             )
        retries = 0
        rdata = None
        while retries < 5:
            try:
                rtype, rdata, rmsgid, serverctrls = l.result3(msgid, timeout=ldap_timeout, resp_ctrl_classes=known_ldap_resp_ctrls)
            except Exception:
                logger.exception('Error getting ldap list memberships')
                retries += 1
            else:
                break
        # ldap sometimees returns good but does not return rdata
        if rdata is None:
            break
        for data in rdata:
            member = data[1].get(search_strings['user_mail_field'], [None])[0]
            if isinstance(member, bytes):
                member = member.decode('utf-8')
            results |= {member}

        pctrls = [c for c in serverctrls
                  if c.controlType == SimplePagedResultsControl.controlType]

        if not pctrls:
            # Paging not supported
            logger.debug('LDAP pagination not supported')
            break
        cookie = pctrls[0].cookie
        if not cookie:
            break
        req_ctrl.cookie = cookie

    return results


def get_ldap_flat_membership(l, search_strings, list_name, max_depth, depth=0, seen_lists=set()):
    members = get_ldap_list_membership(l, search_strings, list_name)

    new_depth = depth + 1
    if new_depth <= max_depth:
        for sub_list, email in get_ldap_lists(l, search_strings, list_name):
            if sub_list in seen_lists:
                logger.debug('avoiding nested list loop with already seen list: %s', sub_list)
                continue
            else:
                seen_lists.add(sub_list)
            sub_list_members = get_ldap_flat_membership(l, search_strings, sub_list, max_depth, new_depth, seen_lists)
            logger.debug('Found %s from %s (depth %s/%s)', len(sub_list_members), email, new_depth, max_depth)
            members |= sub_list_members

    return members


def batch_items_from_list(items, items_per_batch):
    items = list(items)
    pos = 0
    while True:
        new_pos = pos + items_per_batch
        items_this_batch = items[pos:new_pos]
        pos = new_pos
        if not items_this_batch:
            break
        yield items_this_batch


def batch_remove_ldap_memberships(session, list_id, members):
    # Remove these in chunks to avoid a gigantic 'where ID in (..)'
    # query that has thousands of entries.
    deletes_per_match = 50
    for memberships_this_batch in batch_items_from_list(members, deletes_per_match):
        affected = session.execute('''DELETE `mailing_list_membership`
                                      FROM `mailing_list_membership`
                                      JOIN `target_contact` ON `target_contact`.`target_id` = `mailing_list_membership`.`user_id`
                                      WHERE `list_id` = :list_id
                                      AND `target_contact`.`mode_id` = (SELECT `id` FROM `mode` WHERE `name` = 'email')
                                      AND `destination` IN :members''',
                                   {'list_id': list_id, 'members': tuple(memberships_this_batch)}).rowcount
        session.commit()
        sleep(update_sleep)
        logger.info('Deleted %s members from list id %s', affected, list_id)


def sync_ldap_lists(ldap_settings, engine):
    try:
        if 'cert_path' in ldap_settings:
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_ALLOW)
        l = ldap.ldapobject.ReconnectLDAPObject(ldap_settings['connection']['url'])
    except Exception:
        logger.exception('Connecting to ldap to get our mailing lists failed.')
        return

    try:
        if 'cert_path' in ldap_settings:
            l.set_option(ldap.OPT_X_TLS_CACERTFILE, ldap_settings['cert_path'])
        l.simple_bind_s(*ldap_settings['connection']['bind_args'])
    except Exception:
        logger.exception('binding to ldap to get our mailing lists failed.')
        return

    session = sessionmaker(bind=engine)()

    mailing_list_type_name = 'mailing-list'

    list_type_id = session.execute('SELECT `id` FROM `target_type` WHERE `name` = :name', {'name': mailing_list_type_name}).scalar()
    if not list_type_id:
        try:
            list_type_id = session.execute('INSERT INTO `target_type` (`name`) VALUES (:name)', {'name': mailing_list_type_name}).lastrowid
            session.commit()
            logger.info('Created target_type "%s" with id %s', mailing_list_type_name, list_type_id)
        except (IntegrityError, DataError):
            logger.exeption('Failed creating mailing-list type ID')
            return

    ldap_add_pause_interval = ldap_settings.get('user_add_pause_interval', None)
    ldap_add_pause_duration = ldap_settings.get('user_add_pause_duration', 1)

    ldap_lists = get_ldap_lists(l, ldap_settings['search_strings'])
    ldap_lists_count = len(ldap_lists)
    metrics.set('ldap_lists_found', ldap_lists_count)
    metrics.set('ldap_memberships_found', 0)
    logger.info('Found %s ldap lists', ldap_lists_count)

    existing_ldap_lists = {row[0] for row in session.execute('''SELECT `name` FROM `target` WHERE `target`.`type_id` = :type_id''', {'type_id': list_type_id})}
    existing_ldap_lower = {name.lower() for name in existing_ldap_lists}
    existing_case_sensitive_map = {name.lower(): name for name in existing_ldap_lists}
    incoming_case_sensitive_map = {item[1].lower(): item[1] for item in ldap_lists}
    kill_lists = existing_ldap_lower - {item[1].lower() for item in ldap_lists}
    if kill_lists:
        metrics.incr('ldap_lists_removed', len(kill_lists))
        for ldap_list in kill_lists:
            prune_target(engine, existing_case_sensitive_map[ldap_list], mailing_list_type_name)

    # if only case changed rename the mailing list
    case_sensitive_kill = existing_ldap_lists - {item[1] for item in ldap_lists}
    rename_list = {name.lower() for name in case_sensitive_kill} - kill_lists
    if rename_list:
        for ldap_list in rename_list:
            old_name = existing_case_sensitive_map[ldap_list]
            new_name = incoming_case_sensitive_map[ldap_list]
            list_id = session.execute('''SELECT `target`.`id` FROM `target` WHERE `target`.`name` = :name
                                       AND type_id IN ( SELECT `target_type`.`id` FROM `target_type` WHERE `name` = "mailing-list")''', {'name': old_name}).scalar()
            session.execute('UPDATE `mailing_list` SET `name` = :name WHERE `target_id` = :list_id', {'name': new_name, 'list_id': list_id})
            session.commit()
            logger.info('Renamed ldap list %s to %s', old_name, new_name)

    user_add_count = 0

    for list_cn, list_name in ldap_lists:
        try:
            members = get_ldap_flat_membership(l, ldap_settings['search_strings'], list_cn, ldap_settings['max_depth'], 0, set())
        except ldap.SERVER_DOWN:
            # reconnect and retry once
            metrics.incr('ldap_reconnects')
            logger.warning('LDAP server went away for list %s. Reconnecting', list_name)
            l.reconnect(ldap_settings['connection']['url'])
            try:
                members = get_ldap_flat_membership(l, ldap_settings['search_strings'], list_cn, ldap_settings['max_depth'], 0, set())
            except ldap.SERVER_DOWN:
                logger.warning('Reconnect failed for ldap list %s, skipping', list_name)
                continue

        if not members:
            logger.info('Ignoring/pruning empty ldap list %s', list_name)
            continue

        num_members = len(members)
        metrics.incr('ldap_memberships_found', num_members)

        created = False

        list_id = session.execute('''SELECT `target`.`id` FROM `target` WHERE `target`.`name` = :name
                                       AND type_id IN ( SELECT `target_type`.`id` FROM `target_type` WHERE `name` = "mailing-list")''', {'name': list_name}).scalar()

        if not list_id:
            try:
                list_id = session.execute('''INSERT INTO `target` (`type_id`, `name`)
                                             VALUES (:type_id, :name)''', {'type_id': list_type_id, 'name': list_name}).lastrowid
                session.commit()
            except (IntegrityError, DataError):
                logger.exception('Failed adding row to target table for mailing list %s. Skipping this list.', list_name)
                metrics.incr('ldap_lists_failed_to_add')
                continue

            try:
                session.execute('''INSERT INTO `mailing_list` (`target_id`, `count`) VALUES (:list_id, :count)''', {'list_id': list_id, 'count': num_members})
                session.commit()
            except (IntegrityError, DataError):
                logger.exception('Failed adding row to mailing_list table for mailing list %s (ID: %s). Skipping this list.', list_name, list_id)
                metrics.incr('ldap_lists_failed_to_add')
                continue

            logger.info('Created list %s with id %s', list_name, list_id)
            metrics.incr('ldap_lists_added')
            created = True

        if not created:
            session.execute('UPDATE `mailing_list` SET `count` = :count WHERE `target_id` = :list_id', {'count': num_members, 'list_id': list_id})
            session.commit()

        existing_members = {row[0] for row in session.execute('''
                            SELECT `target_contact`.`destination`
                            FROM `mailing_list_membership`
                            JOIN `target_contact` ON `target_contact`.`target_id` = `mailing_list_membership`.`user_id`
                            WHERE `mailing_list_membership`.`list_id` = :list_id
                            AND `target_contact`.`mode_id` = (SELECT `id` FROM `mode` WHERE `name` = 'email')
                            ''', {'list_id': list_id})}

        add_members = members - existing_members
        kill_members = existing_members - members

        if add_members:
            metrics.incr('ldap_memberships_added', len(add_members))

            for member in add_members:
                try:
                    user_id = session.execute('''SELECT `target_id` FROM `target_contact`
                                                JOIN `target` ON `target`.`id` = `target_id`
                                                WHERE `destination` = :name
                                                AND `mode_id` = (SELECT `id` FROM `mode` WHERE `name` = 'email')
                                                AND `target`.`type_id` = (SELECT `id` FROM `target_type` WHERE `name` = 'user')''', {'name': member}).scalar()
                    if user_id is None:
                        continue
                    session.execute('''INSERT INTO `mailing_list_membership` (`list_id`, `user_id`) VALUES (:list_id,:user_id)
                                    ''', {'list_id': list_id, 'user_id': user_id})
                    session.commit()
                    logger.info('Added %s to list %s', member, list_name)
                    user_add_count += 1
                except (IntegrityError, DataError):
                    metrics.incr('ldap_memberships_failed_to_add')
                    logger.warning('Failed adding %s to %s', member, list_name)

                if (ldap_add_pause_interval is not None) and (user_add_count % ldap_add_pause_interval) == 0:
                    logger.info('Pausing for %s seconds every %s users.', ldap_add_pause_duration, ldap_add_pause_interval)
                    time.sleep(ldap_add_pause_duration)
                elif ldap_add_pause_interval is None:
                    sleep(update_sleep)

        if kill_members:
            metrics.incr('ldap_memberships_removed', len(kill_members))
            batch_remove_ldap_memberships(session, list_id, kill_members)
    session.close()


def main():
    global ldap_timeout
    global ldap_pagination_size
    global update_sleep
    config = load_config()
    metrics.init(config, 'iris-sync-targets', stats_reset)

    default_ldap_timeout = 60
    default_ldap_pagination_size = 400
    default_update_sleep = 0
    default_nap_time = 3600

    ldap_timeout = int(config.get('sync_script_ldap_timeout', default_ldap_timeout))
    ldap_pagination_size = int(config.get('sync_script_ldap_pagination_size', default_ldap_pagination_size))
    update_sleep = float(config.get('target_update_pause', default_update_sleep))
    try:
        nap_time = int(config.get('sync_script_nap_time', default_nap_time))
    except ValueError:
        nap_time = default_nap_time

    # check if we are using special connection settings for this script
    if config.get('db_target_sync'):
        engine = create_engine(config['db_target_sync']['conn']['str'] % config['db_target_sync']['conn']['kwargs'],
                               **config['db_target_sync']['kwargs'])
    else:
        engine = create_engine(config['db']['conn']['str'] % config['db']['conn']['kwargs'],
                               **config['db']['kwargs'])

    # Optionally, maintain an internal list of mailing lists from ldap that can also be
    # used as targets.
    ldap_lists = config.get('ldap_lists')

    # Initialize these to zero at the start of the app, and don't reset them at every
    # metrics interval
    metrics.set('users_found', 0)
    metrics.set('teams_found', 0)

    metrics.set('ldap_lists_found', 0)
    metrics.set('ldap_memberships_found', 0)

    metrics_task = spawn(metrics.emit_forever)

    while True:
        if not bool(metrics_task):
            logger.error('metrics task failed, %s', metrics_task.exception)
            metrics_task = spawn(metrics.emit_forever)
        try:
            sync_from_oncall(config, engine)
        except Exception:
            logger.exception('Error syncing from oncall!')

        # Do ldap mailing list sync *after* we do the normal sync, to ensure we have the users
        # which will be in ldap already populated.
        if ldap_lists:

            if 'ldap_cert_path' in ldap_lists:
                ldap_cert_path = ldap_lists['ldap_cert_path']
                if not os.access(ldap_cert_path, os.R_OK):
                    logger.error("Failed to read ldap_cert_path certificate")
                    raise IOError
                else:
                    ldap_lists['cert_path'] = ldap_cert_path
            try:
                list_run_start = time.time()
                sync_ldap_lists(ldap_lists, engine)
                logger.info('Ldap mailing list sync took %.2f seconds', time.time() - list_run_start)
            except Exception:
                logger.exception('Error syncing from ldap!')

        logger.info('Sleeping for %d seconds' % nap_time)
        sleep(nap_time)


if __name__ == '__main__':
    main()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from .. import db
import logging

logger = logging.getLogger(__name__)


def update_message_mode(message):
    if not message.get('message_id'):
        logger.warning('Cannot change mode for message %s as it is missing message_id', message)
        return

    session = db.Session()
    mode_id = session.execute('SELECT `id` FROM `mode` WHERE `name` = :mode', message).scalar()

    # Need to update mode_id in the dictionary as its gets set in DB in other parts of the sender
    if mode_id:
        message['mode_id'] = mode_id
        session.execute('UPDATE `message` SET `mode_id` = :mode_id WHERE `id` = :message_id', message)
        session.commit()
    else:
        logger.warning('Cannot update mode for message %(message_id)s to %(mode)s as looking up its ID failed', message)

    session.close()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

"""Tests for iris."""
import falcon.testing
import iris.cache
from iris.api import ReqBodyMiddleware, AuthMiddleware, Healthcheck
import time
import hmac
import hashlib
import base64
from mock import patch, mock_open


class TestCommand(falcon.testing.TestCase):
    def test_parse_response(self):
        from iris.utils import parse_response
        with patch('iris.utils.db'):
            msg_id, cmd = parse_response('123 Claim', 'sms', '123-456-7890')
            self.assertEqual(msg_id, '123')
            self.assertEqual(cmd, 'claim')

            msg_id, cmd = parse_response('Claim 123', 'sms', '123-456-7890')
            self.assertEqual(msg_id, '123')
            self.assertEqual(cmd, 'claim')

            msg_id, cmd = parse_response('claim 123', 'sms', '123-456-7890')
            self.assertEqual(msg_id, '123')
            self.assertEqual(cmd, 'claim')


class TestHealthcheck(falcon.testing.TestCase):
    def test_healthcheck(self):
        with patch('iris.api.open', mock_open(read_data='GOOD')) as m:
            self.api.add_route('/healthcheck', Healthcheck('healthcheck_path'))
            result = self.simulate_get(path='/healthcheck')
            m.assert_called_once_with('healthcheck_path')
        self.assertEqual(result.status_code, 503)
        self.assertEqual(result.content, b'Could not connect to database')


class TestAuth(falcon.testing.TestCase):

    class DummyResource(object):
        allow_read_no_auth = False

        def on_get(self, req, resp):
            resp.status = falcon.HTTP_200
            resp.body = 'Hello world'

    def test_auth(self):
        iris.cache.applications = {'app': {'key': 'key', 'secondary_key': None}}
        api = falcon.API(middleware=[ReqBodyMiddleware(), AuthMiddleware()])
        dummy = self.DummyResource()
        api.add_route('/foo/bar', dummy)
        self.api = api

        window = int(time.time()) // 5
        text = '%s %s %s %s' % (window, 'GET', '/foo/bar', '')
        HMAC = hmac.new(b'key', text.encode('utf-8'), hashlib.sha512)
        digest = base64.urlsafe_b64encode(HMAC.digest()).decode('utf-8')
        auth = 'hmac app:%s' % digest
        result = self.simulate_get(path='/foo/bar', headers={'Authorization': auth})
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')

        # Test query string
        window = int(time.time()) // 5
        text = '%s %s %s %s' % (window, 'GET', '/foo/bar?baz=123', '')
        HMAC = hmac.new(b'key', text.encode('utf-8'), hashlib.sha512)
        digest = base64.urlsafe_b64encode(HMAC.digest()).decode('utf-8')
        auth = 'hmac app:%s' % digest
        result = self.simulate_get(path='/foo/bar', query_string='baz=123', headers={'Authorization': auth})
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')

        # Test trailng slash
        window = int(time.time()) // 5
        text = '%s %s %s %s' % (window, 'GET', '/foo/bar/', '')
        HMAC = hmac.new(b'key', text.encode('utf-8'), hashlib.sha512)
        digest = base64.urlsafe_b64encode(HMAC.digest()).decode('utf-8')
        auth = 'hmac app:%s' % digest
        result = self.simulate_get(path='/foo/bar/', headers={'Authorization': auth})
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')

        # Test no auth header
        result = self.simulate_get(path='/foo/bar')
        self.assertEqual(result.status_code, 401)

        # Test bad auth header
        result = self.simulate_get(path='/foo/bar',
                                   headers={'Authorization': 'foo' + auth})
        self.assertEqual(result.status_code, 401)

        # Test read only
        dummy.allow_read_no_auth = True
        result = self.simulate_get(path='/foo/bar')
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')

    def test_secondary_auth(self):
        iris.cache.applications = {'app': {'key': 'asdf', 'secondary_key': 'key'}}
        api = falcon.API(middleware=[ReqBodyMiddleware(), AuthMiddleware()])
        dummy = self.DummyResource()
        api.add_route('/foo/bar', dummy)
        self.api = api

        window = int(time.time()) // 5
        text = '%s %s %s %s' % (window, 'GET', '/foo/bar', '')
        HMAC = hmac.new(b'key', text.encode('utf-8'), hashlib.sha512)
        digest = base64.urlsafe_b64encode(HMAC.digest()).decode('utf-8')
        auth = 'hmac app:%s' % digest
        result = self.simulate_get(path='/foo/bar', headers={'Authorization': auth})
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')

        # Test query string
        window = int(time.time()) // 5
        text = '%s %s %s %s' % (window, 'GET', '/foo/bar?baz=123', '')
        HMAC = hmac.new(b'key', text.encode('utf-8'), hashlib.sha512)
        digest = base64.urlsafe_b64encode(HMAC.digest()).decode('utf-8')
        auth = 'hmac app:%s' % digest
        result = self.simulate_get(path='/foo/bar', query_string='baz=123', headers={'Authorization': auth})
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')

        # Test trailng slash
        window = int(time.time()) // 5
        text = '%s %s %s %s' % (window, 'GET', '/foo/bar/', '')
        HMAC = hmac.new(b'key', text.encode('utf-8'), hashlib.sha512)
        digest = base64.urlsafe_b64encode(HMAC.digest()).decode('utf-8')
        auth = 'hmac app:%s' % digest
        result = self.simulate_get(path='/foo/bar/', headers={'Authorization': auth})
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')

        # Test no auth header
        result = self.simulate_get(path='/foo/bar')
        self.assertEqual(result.status_code, 401)

        # Test bad auth header
        result = self.simulate_get(path='/foo/bar',
                                   headers={'Authorization': 'foo' + auth})
        self.assertEqual(result.status_code, 401)

        # Test read only
        dummy.allow_read_no_auth = True
        result = self.simulate_get(path='/foo/bar')
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content, b'Hello world')



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from iris.vendors.iris_hipchat import iris_hipchat


def test_message_construction_for_incident():
    hipchat_vendor = iris_hipchat({
        'auth_token': 'abc',
        'iris_incident_url': 'http://foo.bar/incidents',
        'room_id': 1234,
    })
    fake_msg = {
        'application': 'grafana',
        'incident_id': 123,
        'body': 'test body',
        'message_id': 456,
        'destination': '@user1',
    }
    msg_payload = hipchat_vendor.get_message_payload(fake_msg, "@user1")
    assert msg_payload['message'] == '@user1 %s' % fake_msg['body']


def test_destination_parsing_for_incident():
    hipchat_vendor = iris_hipchat({
        'auth_token': 'abc',
        'iris_incident_url': 'http://foo.bar/incidents',
        'room_id': 1234,
    })
    fake_msg = {
        'application': 'grafana',
        'incident_id': 123,
        'body': 'test body',
        'message_id': 456,
    }
    destination = '1234;testtoken;@user1'
    fake_msg['destination'] = destination
    room_id, token, mention = hipchat_vendor.parse_destination(fake_msg['destination'])

    assert room_id == 1234
    assert token == 'testtoken'
    assert mention == '@user1'

    destination = '1234;testtoken'
    fake_msg['destination'] = destination
    room_id, token, mention = hipchat_vendor.parse_destination(fake_msg['destination'])

    assert room_id == 1234
    assert token == 'testtoken'
    assert mention == ''


def test_destination_parsing_defaults_for_incident():
    hipchat_vendor = iris_hipchat({
        'auth_token': 'validtoken',
        'iris_incident_url': 'http://foo.bar/incidents',
        'room_id': 1234,
    })
    fake_msg = {
        'application': 'grafana',
        'incident_id': 123,
        'body': 'test body',
        'message_id': 456,
    }
    destination = 'user_missing_@'
    fake_msg['destination'] = destination
    room_id, token, mention = hipchat_vendor.parse_destination(fake_msg['destination'])

    assert room_id == 1234
    assert token == 'validtoken'
    assert mention == ''

    destination = 'not_an_int;testtoken'
    fake_msg['destination'] = destination
    room_id, token, mention = hipchat_vendor.parse_destination(fake_msg['destination'])

    assert room_id == 1234
    assert token == 'testtoken'
    assert mention == ''



from gevent import monkey, sleep
monkey.patch_all()  # NOQA

from iris import metrics
from iris.config import load_config
from iris.client import IrisClient
import exchangelib
import exchangelib.errors
import exchangelib.protocol
import sys
import logging
import logging.handlers
import os
import time
import requests


logger = logging.getLogger()
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s')
log_file = os.environ.get('OWA_SYNC_LOG_FILE')
if log_file:
    ch = logging.handlers.RotatingFileHandler(log_file, mode='a', maxBytes=10485760, backupCount=10)
else:
    ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.setLevel(logging.INFO)
logger.addHandler(ch)

pidfile = os.environ.get('OWA_SYNC_PIDFILE')
if pidfile:
    try:
        pid = os.getpid()
        with open(pidfile, 'w') as h:
            h.write('%s\n' % pid)
            logger.info('Wrote pid %s to %s', pid, pidfile)
    except IOError:
        logger.exception('Failed writing pid to %s', pidfile)

default_metrics = {
    'owa_api_failure_count': 0,  # we fail to hit EWS api. requests timeout or similar
    'message_relay_success_count': 0,  # iris-api gives 2xx
    'message_relay_failure_count': 0,  # iris-api gives 5xx
    'malformed_message_count': 0,  # iris-api gives 4xx (likely not fault, so don't blame ourselves)
    'total_inbox_count': 0,
    'unread_inbox_count': 0,
    'message_process_count': 0,
    'message_ignore_count': 0,
    'incident_created_count': 0,
}


email_headers_to_ignore = frozenset([('X-Autoreply', 'yes'),
                                     ('Auto-Submitted', 'auto-replied'),
                                     ('Precedence', 'bulk')])


# per exchangelib docs, to customize the http requests sent (eg to add a proxy)
# need to create a custom requests adapter
class UseProxyHttpAdapter(requests.adapters.HTTPAdapter):
    _my_proxies = None

    def send(self, *args, **kwargs):
        if self._my_proxies:
            kwargs['proxies'] = self._my_proxies
        return super(UseProxyHttpAdapter, self).send(*args, **kwargs)


def split_in_batches(l, n):
    # For item i in a range that is a length of l,
    for i in range(0, len(l), n):
        # Create an index range for l of n items:
        yield l[i:i + n]


def is_pointless_message(headers):
    for header in email_headers_to_ignore:
        if {'name': header[0], 'value': header[1]} in headers:
            logger.warning('Filtering out message due to header combination %s: %s', *header)
            return True
    return False


def poll(account, iris_client):

    try:
        metrics.set('total_inbox_count', account.inbox.total_count)
        metrics.set('unread_inbox_count', account.inbox.unread_count)
    except (exchangelib.errors.EWSError, requests.exceptions.RequestException):
        logger.exception('Failed to gather inbox counts from OWA API')
        metrics.incr('owa_api_failure_count')

    processed_messages = 0
    messages_to_mark_read = []
    messages_to_delete = []

    try:
        for message in account.inbox.filter(is_read=False).order_by('-datetime_received'):
            processed_messages += 1

            try:
                result = relay(message, iris_client)
                if not result:
                    messages_to_delete.append(message)
            except Exception:
                logger.exception('Uncaught exception during message relaying')
                metrics.incr('message_relay_failure_count')

            # Mark it as read in bulk later. (This syntax isn't documented)
            message.is_read = True
            messages_to_mark_read.append((message, ('is_read', )))

    except (exchangelib.errors.EWSError, requests.exceptions.RequestException):
        logger.exception('Failed to iterate through inbox')
        metrics.incr('owa_api_failure_count')

    if messages_to_mark_read:
        # mark as read in batches of 100 messages
        messages_in_batches = split_in_batches(messages_to_mark_read, 100)
        for batch in messages_in_batches:
            bulk_update_count = len(batch)
            logger.info('will mark %s messages as read', bulk_update_count)
            try:
                account.bulk_update(items=batch)
            except (exchangelib.errors.EWSError, requests.exceptions.RequestException):
                logger.exception('Failed to update read status on %s messages in bulk', bulk_update_count)
                metrics.incr('owa_api_failure_count')

    if messages_to_delete:
        for batch in split_in_batches(messages_to_delete, 100):
            bulk_delete_count = len(batch)
            logger.info('will delete %s messages', bulk_delete_count)
            try:
                account.bulk_delete(ids=batch)
            except (exchangelib.errors.EWSError, requests.exceptions.RequestException):
                logger.exception('Failed to delete on %s messages in bulk', bulk_delete_count)
                metrics.incr('owa_api_failure_count')

    metrics.set('message_process_count', processed_messages)
    return processed_messages


def relay(message, iris_client):
    if message.headers is None:
        logger.info('Ignoring message with no headers %s (from %s)', message.message_id, message.sender.email_address)
        metrics.incr('message_ignore_count')
        return True

    # Get headers into the format the iris expects from gmail
    headers = [{'name': header.name, 'value': header.value} for header in message.headers]

    # If this is a bulk message or an auto reply or something else, don't bother sending it to iris-api
    if is_pointless_message(headers):
        logger.info('Not relaying pointless message %s (from %s) to iris-api', message.message_id, message.sender.email_address)
        metrics.incr('message_ignore_count')
        return True

    # To and From headers are strangely missing
    if message.to_recipients:
        headers.append({'name': 'To', 'value': [r.email_address for r in message.to_recipients]})
    headers.append({'name': 'From', 'value': message.sender.email_address})

    data = {'headers': headers, 'body': message.text_body.strip()}

    try:
        req = iris_client.post('response/email', json=data)
    except requests.exceptions.RequestException:
        metrics.incr('message_relay_failure_count')
        logger.exception('Failed posting message %s (from %s) to iris-api', message.message_id, message.sender.email_address)
        return False

    result = False
    code_type = req.status_code // 100

    if code_type == 5:
        metrics.incr('message_relay_failure_count')
        logger.error('Failed posting message %s (from %s) to iris-api. Got status code %s and response %s',
                     message.message_id, message.sender.email_address, req.status_code, req.text)

    elif code_type == 4:
        metrics.incr('malformed_message_count')
        logger.error('Failed posting message %s (from %s) to iris-api. Message likely malformed. Status code: %s. Response: %s',
                     message.message_id, message.sender.email_address, req.status_code, req.text)

    elif code_type == 2:
        metrics.incr('message_relay_success_count')

        # If we create an incident using an email, this header will be set and will be the numeric ID
        # of the created incident; otherwise, the header will not exist or it will be a textual
        # error message.
        incident_header = req.headers.get('X-IRIS-INCIDENT')
        if isinstance(incident_header, str) and incident_header.isdigit():
            metrics.incr('incident_created_count')
            result = True

    else:
        logger.error('Failed posting message %s (from %s) to iris-api. Message likely malformed. Got back strange status code: %s. Response: %s',
                     message.message_id, message.sender.email_address, req.status_code, req.text)

    return result


def main():
    boot_time = time.time()
    config = load_config()

    metrics.init(config, 'iris-owa-sync', default_metrics)

    owaconfig = config.get('owa')

    if not owaconfig:
        logger.critical('Missing OWA configs')
        sys.exit(1)

    api_host = owaconfig.get('api_host', 'http://localhost:16649')
    iris_client = IrisClient(api_host, 0, owaconfig['iris_app'], owaconfig['iris_app_key'])

    proxies = owaconfig.get('proxies')

    # only way to configure a proxy is to monkey-patch (http adapter) a monkey-patch (baseprotocol) :/
    if proxies:
        UseProxyHttpAdapter._my_proxies = proxies
        exchangelib.protocol.BaseProtocol.HTTP_ADAPTER_CLS = UseProxyHttpAdapter

    creds = exchangelib.Credentials(**owaconfig['credentials'])

    try:
        nap_time = int(owaconfig.get('sleep_interval', 60))
    except ValueError:
        nap_time = 60

    while True:
        start_time = time.time()
        message_count = 0

        try:
            config = exchangelib.Configuration(credentials=creds, **owaconfig['config'])
            account = exchangelib.Account(
                config=config,
                access_type=exchangelib.DELEGATE,
                **owaconfig['account'])
        except (exchangelib.errors.EWSError, requests.exceptions.RequestException):
            logger.exception('Failed authenticating to OWA365')
            metrics.incr('owa_api_failure_count')
        else:
            logger.info('Receiving mail on behalf of %s', owaconfig['account'].get('primary_smtp_address'))
            message_count = poll(account, iris_client)

        now = time.time()
        run_time = now - start_time
        logger.info('Last run took %2.f seconds and processed %s messages. Waiting %s seconds until next poll..', run_time, message_count, nap_time)
        metrics.set('uptime', now - boot_time)
        metrics.emit()
        sleep(nap_time)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris.constants import SMS_SUPPORT, CALL_SUPPORT
from iris.plugins import find_plugin
from iris.custom_import import import_custom_module
from twilio.rest import Client
from twilio.http.http_client import TwilioHttpClient
from iris import db
from sqlalchemy.exc import IntegrityError
import time
import urllib.parse
import logging

logger = logging.getLogger(__name__)


class TwilioHTTP(TwilioHttpClient):
    def __init__(self, timeout):
        self.timeout = timeout
        super().__init__()

    def request(self, method, url, params=None, data=None, headers=None, auth=None,
                timeout=None, allow_redirects=False):
        return super().request(method, url, params=params, data=data, headers=headers, auth=auth,
                               timeout=self.timeout, allow_redirects=allow_redirects)


class iris_twilio(object):
    supports = frozenset([SMS_SUPPORT, CALL_SUPPORT])

    def __init__(self, config):
        self.config = config
        self.http_client = TwilioHTTP(self.config.get('timeout', 10))
        if 'proxy' in self.config:
            proxy = '%s:%s' % (self.config['proxy']['host'], self.config['proxy']['port'])
            self.http_client.session.proxies = {'https': proxy, 'http': proxy}
        self.modes = {
            SMS_SUPPORT: self.send_sms,
            CALL_SUPPORT: self.send_call,
        }
        self.say_endpoint = config.get('say_endpoint', '/api/v0/twilio/calls/say')
        self.gather_endpoint = config.get('gather_endpoint', '/api/v0/twilio/calls/gather')
        push_config = config.get('push_notification', {})
        self.push_active = push_config.get('activated', False)
        if self.push_active:
            self.notifier = import_custom_module('iris.push', push_config['type'])(push_config)

    def get_twilio_client(self):
        return Client(self.config['account_sid'],
                      self.config['auth_token'],
                      http_client=self.http_client)

    def generate_message_text(self, message):
        content = []

        for key in ('subject', 'body'):
            value = message.get(key)
            if not isinstance(value, str):
                continue
            value = value.strip()
            if value:
                content.append(value)

        return '. '.join(content)

    def initialize_twilio_message_status(self, sid, message_id):
        session = db.Session()
        try:
            session.execute('''INSERT INTO `twilio_delivery_status` (`twilio_sid`, `message_id`)
                               VALUES (:sid, :mid)''', {'sid': sid, 'mid': message_id})
            session.commit()
        except IntegrityError:
            logger.exception('Failed initializing twilio delivery status row. sid: %s, mid: %s', sid, message_id)
        session.close()

    def send_sms(self, message):
        client = self.get_twilio_client()

        sender = client.messages.create
        from_ = self.config['twilio_number']
        start = time.time()
        content = self.generate_message_text(message)
        status_callback_url = self.config['relay_base_url'] + '/api/v0/twilio/status'
        result = sender(to=message['destination'],
                        from_=from_,
                        body=content[:1600],
                        status_callback=status_callback_url)

        send_time = time.time() - start

        try:
            sid = result.sid
        except Exception:
            logger.exception('Failed getting Message SID from Twilio')
            sid = None

        message_id = message.get('message_id')

        if sid and message_id:
            self.initialize_twilio_message_status(sid, message_id)
        else:
            logger.warning('Not initializing twilio SMS status row (mid: %s, sid: %s)', message_id, sid)

        return send_time

    def send_call(self, message):
        plugin = find_plugin(message['application'])
        if not plugin:
            raise ValueError('not supported source: %(application)s' % message)

        client = self.get_twilio_client()
        sender = client.calls.create
        from_ = self.config['twilio_number']
        status_callback_url = self.config['relay_base_url'] + '/api/v0/twilio/status'
        content = self.generate_message_text(message)

        payload = {
            'content': content[:480],
            'loop': 3,
            'source': message['application'],
        }

        # If message_id is None or 0, go through says as iris can't handle
        # phone call response without the id
        message_id = message.get('message_id')
        if message_id:
            payload['message_id'] = message_id
            payload['instruction'] = plugin.get_phone_menu_text()
            relay_cb_url = '%s%s?%s' % (
                self.config['relay_base_url'], self.gather_endpoint, urllib.parse.urlencode({k: str(v).encode('utf-8') for k, v in payload.items()})
            )
        else:
            relay_cb_url = '%s%s?%s' % (
                self.config['relay_base_url'], self.say_endpoint, urllib.parse.urlencode({k: str(v).encode('utf-8') for k, v in payload.items()})
            )

        start = time.time()

        result = sender(to=message['destination'],
                        from_=from_,
                        if_machine='Continue',
                        url=relay_cb_url,
                        status_callback=status_callback_url)

        send_time = time.time() - start

        try:
            sid = result.sid
        except Exception:
            logger.exception('Failed getting Message SID from Twilio')
            sid = None

        if sid and message_id:
            self.initialize_twilio_message_status(sid, message_id)
        else:
            logger.warning('Not initializing twilio call status row (mid: %s, sid: %s)', message_id, sid)

        return send_time

    def send(self, message, customizations=None):
        if self.push_active:
            self.notifier.send_push(message)
        return self.modes[message['mode']](message)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from iris.vendors.iris_slack import iris_slack
import ujson as json


def test_atttachments_construction_for_incident():
    slack_vendor = iris_slack({
        'auth_token': 'abc',
        'iris_incident_url': 'http://foo.bar/incidents',
        'message_attachments': {
            'fallback': 'foo fallback',
            'pretext': 'foo pretext',
        }
    })
    fake_msg = {
        'application': 'grafana',
        'incident_id': 123,
        'body': 'test body',
        'message_id': 456,
        'destination': 'user1'
    }
    msg_payload = slack_vendor.get_message_payload(fake_msg)
    assert msg_payload['text'] == '[grafana] %s' % fake_msg['body']
    assert msg_payload['token'] == 'abc'
    assert msg_payload['channel'] == '@user1'

    attachments = msg_payload['attachments']
    assert json.loads(attachments) == [{
        'fallback': 'foo fallback',
        'pretext': 'foo pretext',
        'title': 'Iris incident %r' % fake_msg['incident_id'],
        'mrkdwn_in': ['pretext'],
        'attachment_type': 'default',
        'callback_id': fake_msg['message_id'],
        'color': 'danger',
        'title_link': 'http://foo.bar/incidents/%d' % fake_msg['incident_id'],
        'actions': [
            {
                'name': 'claim',
                'text': 'Claim Incident',
                'type': 'button',
                'value': 'claimed'
            },
            {
                'name': 'claim all',
                'text': 'Claim All',
                'style': 'danger',
                'type': 'button',
                'value': 'claimed all',
                "confirm": {
                    "title": "Are you sure?",
                    "text": "This will claim all active incidents targeting you.",
                    "ok_text": "Yes",
                    "dismiss_text": "No"
                }
            }
        ]
    }]


def test_atttachments_construction_for_notification():
    slack_vendor = iris_slack({
        'auth_token': 'abc',
        'iris_incident_url': 'http://foo.bar/incidents',
        'message_attachments': {
            'fallback': 'foo fallback',
            'pretext': 'foo pretext',
        }
    })
    fake_msg = {
        'application': 'grafana',
        'body': 'test body notification',
        'destination': 'user1'
    }
    msg_payload = slack_vendor.get_message_payload(fake_msg)
    assert msg_payload == {
        'text': '[grafana] %s' % fake_msg['body'],
        'token': 'abc',
        'channel': '@user1'
    }



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from contextlib import contextmanager
from .validators import IrisValidationException
from falcon import HTTPBadRequest, HTTPNotFound, HTTPForbidden, HTTPUnauthorized
import logging

logger = logging.getLogger(__name__)


Session = None
dict_cursor = None
ss_dict_cursor = None
engine = None


def init(config):
    global engine
    global dict_cursor
    global ss_dict_cursor
    global Session

    engine = create_engine(config['db']['conn']['str'] % config['db']['conn']['kwargs'],
                           **config['db']['kwargs'])
    dict_cursor = engine.dialect.dbapi.cursors.DictCursor
    ss_dict_cursor = engine.dialect.dbapi.cursors.SSDictCursor
    Session = sessionmaker(bind=engine)


@contextmanager
def guarded_session():
    '''
    Context manager that will automatically close session on exceptions
    '''
    try:
        session = Session()
        yield session
    except IrisValidationException as e:
        session.close()
        raise HTTPBadRequest('Validation error', str(e))
    except (HTTPForbidden, HTTPUnauthorized, HTTPNotFound, HTTPBadRequest):
        session.close()
        raise
    except Exception:
        session.close()
        logger.exception('SERVER ERROR')
        raise



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import pytest
import socket
import requests
import re
from iris.metrics import get_metrics_provider

# We might not have the prometheus library installed.
pytest.importorskip('prometheus_client')


# in case multiple of this test is ran at once, (try to) avoid port conflicts,
# by starting on a random port
def get_free_port():
    s = socket.socket()
    s.bind(('127.0.0.1', 0))
    port = s.getsockname()[1]
    s.close()
    return port


listen_port = get_free_port()

config = {
    'metrics': 'prometheus',
    'prometheus': {
        'test': {
            'server_port': listen_port
        }
    }
}


@pytest.fixture(scope='module')
def metrics():
    return get_metrics_provider(config, 'test')


def test_guage(metrics):
    metrics.send_metrics({
        'value1': 100
    })
    data = requests.get('http://localhost:%d/' % listen_port).text
    m = re.search('^test_value1 (\S+)$', data, re.MULTILINE)
    assert m is not None
    assert m.group(1) == '100.0'



#!/usr/bin/env python

# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# -*- coding:utf-8 -*-

import time
from iris.bin.sender import init_sender
from iris.vendors import IrisVendorManager
import msgpack
import gevent.queue


def test_configure(mocker):
    mocker.patch('iris.sender.cache.RoleTargets.initialize_active_targets')
    mocker.patch('iris.db.init')
    mocker.patch('iris.bin.sender.api_cache.cache_priorities')
    mocker.patch('iris.bin.sender.api_cache.cache_applications')
    mocker.patch('iris.bin.sender.api_cache.cache_modes')
    mock_iris_client = mocker.MagicMock()
    mocker.patch('iris.sender.cache.IrisClient').return_value = mock_iris_client
    mock_iris_client.get.return_value.status_code = 200

    init_sender({
        'db': {
            'conn': {
                'kwargs': {
                    'scheme': 'mysql+pymysql',
                    'user': 'foo',
                    'password': 'bar',
                    'host': '127.0.0.1',
                    'database': 'iris',
                    'charset': 'utf8',
                },
                'str': '%(scheme)s://%(user)s:%(password)s@%(host)s/%(database)s?charset=%(charset)s'
            },
            'kwargs': {
                'pool_recycle': 3600,
                'echo': True,
                'pool_size': 100,
                'max_overflow': 100,
                'pool_timeout': 60,
            }
        },
        'sender': {
            'debug': True,
            'skipgmailwatch': True,
        },
        'oncall': 'http://localhost:8002',
        'role_lookup': 'dummy',
        'metrics': 'dummy',
        'skipsend': True,
    })


fake_message = {
    'message_id': 1234,
    'plan_id': 19546,
    'application': 'test-app',
    'priority': 'high',
    'target': 'test-user',
    'mode': 'sms',
}

fake_notification = {
    'application': 'test-app',
    'priority': 'high',
    'target': 'test-user',
    'role': 'user',
    'subject': 'test subject',
    'body': 'test body',
    'mode': 'email',
}

fake_plan = {
    'name': 'find-test-user',
    'threshold_count': 10,
    'creator': 'test-user',
    'created': 1470444636,
    'aggregation_reset': 300,
    'aggregation_window': 300,
    'threshold_window': 900,
    'tracking_type': None,
    'steps': [
        [{'repeat': 0, 'target': 'test-user', 'id': 178243, 'priority': 'low', 'step': 1,
          'role': 'user', 'template': 'test-app Default', 'wait': 0},
         {'repeat': 1, 'target': 'test-user', 'id': 178252, 'priority': 'high', 'step': 1,
          'role': 'user', 'template': 'test-app Default', 'wait': 300}],
        [{'repeat': 3, 'target': 'test-user', 'id': 178261, 'priority': 'urgent', 'step': 2,
          'role': 'user', 'template': 'test-app Default', 'wait': 900}]
    ],
    'tracking_template': None,
    'tracking_key': None,
    'active': 1,
    'id': 19546,
    'description': "please don't abuse this plan :)"
}


def init_queue_with_item(queue, item=None):
    # drain out queue
    while queue.qsize() > 0:
        queue.get()
    if item:
        queue.put(item)


def test_fetch_and_prepare_message(mocker):
    mocker.patch('iris.bin.sender.message_send_enqueue')
    from iris.bin.sender import (
        fetch_and_prepare_message, message_queue, per_mode_send_queues
    )

    init_queue_with_item(message_queue, {'message_id': 1234, 'plan_id': None})
    fetch_and_prepare_message()
    assert message_queue.qsize() == 0

    send_queue = per_mode_send_queues.setdefault('email', gevent.queue.Queue())

    init_queue_with_item(send_queue, {'message_id': 1234, 'plan_id': None})

    assert message_queue.qsize() == 0
    assert send_queue.qsize() == 1
    m = send_queue.get()
    assert m['message_id'] == 1234


def test_fetch_and_send_message(mocker):
    def check_mark_message_sent(m):
        assert m['message_id'] == fake_message['message_id']

    def mock_set_target_contact(message):
        message['destination'] = 'foo@example.com'
        message['mode'] = 'email'
        message['mode_id'] = 1
        return True

    vendors = IrisVendorManager({}, [])
    mocker.patch('iris.bin.sender.db')
    mocker.patch('iris.bin.sender.distributed_send_message').return_value = True, True
    mocker.patch('iris.bin.sender.quota')
    mocker.patch('iris.metrics.stats')
    mocker.patch('iris.bin.sender.update_message_mode')
    mock_mark_message_sent = mocker.patch('iris.bin.sender.mark_message_as_sent')
    mock_mark_message_sent.side_effect = check_mark_message_sent
    mocker.patch('iris.bin.sender.set_target_contact').side_effect = mock_set_target_contact
    mock_iris_client = mocker.patch('iris.sender.cache.iris_client')
    mock_iris_client.get.return_value.json.return_value = fake_plan
    from iris.bin.sender import (
        fetch_and_send_message, per_mode_send_queues
    )

    send_queue = per_mode_send_queues.setdefault('email', gevent.queue.Queue())

    # drain out send queue
    while send_queue.qsize() > 0:
        send_queue.get()
    send_queue.put(fake_message)

    fetch_and_send_message(send_queue, vendors)

    assert send_queue.qsize() == 0
    mock_mark_message_sent.assert_called_once()


def test_message_retry(mocker):
    def check_mark_message_sent(m):
        assert m['message_id'] == fake_message['message_id']

    def mock_set_target_contact(message):
        message['destination'] = 'foo@example.com'
        message['mode'] = 'sms'
        message['mode_id'] = 1
        return True

    vendors = IrisVendorManager({}, [])
    mocker.patch('iris.bin.sender.db')
    mock_distributed_send_message = mocker.patch('iris.bin.sender.distributed_send_message')
    mock_distributed_send_message.return_value = (False, True)
    mocker.patch('iris.bin.sender.quota')
    mocker.patch('iris.bin.sender.update_message_mode')
    mock_mark_message_sent = mocker.patch('iris.bin.sender.mark_message_as_sent')
    mock_mark_message_sent.side_effect = check_mark_message_sent
    mocker.patch('iris.bin.sender.set_target_contact').side_effect = mock_set_target_contact
    mocker.patch('iris.bin.sender.set_target_fallback_mode').side_effect = mock_set_target_contact
    mock_iris_client = mocker.patch('iris.sender.cache.iris_client')
    mock_iris_client.get.return_value.json.return_value = fake_plan
    from iris.bin.sender import (
        fetch_and_send_message, per_mode_send_queues, metrics, default_sender_metrics,
        add_mode_stat
    )

    def fail_send_message(message, vendors=None):
        add_mode_stat(message['mode'], None)
        return False, True

    mock_distributed_send_message.side_effect = fail_send_message

    metrics.stats.update(default_sender_metrics)

    send_queue = per_mode_send_queues.setdefault('sms', gevent.queue.Queue())

    # drain out send queue
    while send_queue.qsize() > 0:
        send_queue.get()
    send_queue.put(fake_message.copy())

    mock_distributed_send_message.reset_mock()
    fetch_and_send_message(send_queue, vendors)
    mock_distributed_send_message.assert_called()

    # will retry first time
    assert send_queue.qsize() == 1
    mock_mark_message_sent.assert_not_called()

    mock_distributed_send_message.reset_mock()
    fetch_and_send_message(send_queue, vendors)
    mock_distributed_send_message.assert_called()

    # will retry a 2nd time
    assert send_queue.qsize() == 1
    mock_mark_message_sent.assert_not_called()

    mock_distributed_send_message.reset_mock()
    fetch_and_send_message(send_queue, vendors)

    # will not retry a 3rd time
    assert send_queue.qsize() == 0
    mock_distributed_send_message.assert_not_called()

    # we retried after it failed the first time
    assert metrics.stats['message_retry_cnt'] == 2
    assert metrics.stats['sms_fail'] == 2


def test_no_valid_modes(mocker):
    def check_mark_message_sent(m):
        assert m['message_id'] == fake_message['message_id']

    def mock_set_target_contact(message):
        return False

    mocker.patch('iris.metrics.stats')
    mocker.patch('iris.bin.sender.db')
    mocker.patch('iris.bin.sender.set_target_contact').return_value = False
    mock_mark_message_no_contact = mocker.patch('iris.bin.sender.mark_message_has_no_contact')
    mock_mark_message_no_contact.reset_mock()
    from iris.bin.sender import message_send_enqueue, message_ids_being_sent
    message_ids_being_sent.clear()

    message_send_enqueue(fake_message)
    mock_mark_message_no_contact.assert_called_once()


def test_handle_api_request_v0_send(mocker):
    from iris.bin.sender import message_send_enqueue
    from iris.sender.rpc import handle_api_request, send_funcs
    from iris.sender.shared import per_mode_send_queues

    send_funcs['message_send_enqueue'] = message_send_enqueue

    send_queue = per_mode_send_queues.setdefault('email', gevent.queue.Queue())

    # support expanding target
    mocker.patch('iris.sender.cache.targets_for_role', lambda role, target: [target])
    mocker.patch('iris.bin.sender.db')
    mocker.patch('iris.metrics.stats')
    mocker.patch('iris.bin.sender.set_target_contact').return_value = True

    mock_address = mocker.MagicMock()
    mock_socket = mocker.MagicMock()
    mock_socket.recv.return_value = msgpack.packb({
        'endpoint': 'v0/send',
        'data': fake_notification,
    })

    while send_queue.qsize() > 0:
        send_queue.get()

    handle_api_request(mock_socket, mock_address)

    assert send_queue.qsize() == 1
    m = send_queue.get()
    assert m['subject'] == '[%s] %s' % (fake_notification['application'],
                                        fake_notification['subject'])


def test_handle_api_request_v0_send_timeout(mocker):
    import iris.sender.rpc
    iris.sender.rpc.rpc_timeout = 5

    def slee_10(x):
        from gevent import sleep
        sleep(10)

    mock_address = mocker.MagicMock()
    mock_socket = mocker.MagicMock()
    mock_socket.recv.side_effect = slee_10

    iris.sender.rpc.handle_api_request(mock_socket, mock_address)

    mock_socket.sendall.assert_called_with(msgpack.packb('TIMEOUT'))


def test_render_email_response_message(mocker):
    from iris.bin.sender import render
    mock_cursor = mocker.MagicMock()
    mock_db = mocker.patch('iris.bin.sender.db')
    mock_db.engine.raw_connection().cursor.return_value = mock_cursor
    mock_cursor.fetchone.return_value = ['bar', 'foo']

    mock_message = {'message_id': 1}
    render(mock_message)

    mock_cursor.execute.assert_called_once_with('SELECT `body`, `subject` FROM `message` WHERE `id` = %s', 1)
    assert mock_message['body'] == 'bar'
    assert mock_message['subject'] == 'foo'


def test_msgpack_handle_sets():
    from iris.sender.rpc import msgpack_handle_sets
    assert msgpack_handle_sets(set([1, 2, 3, 4])) == [1, 2, 3, 4]


def test_generate_slave_message_payload():
    from iris.sender.rpc import generate_msgpack_message_payload
    data = {
        'ids': set([1, 2, 3, 4])
    }
    result = generate_msgpack_message_payload(data)
    assert msgpack.unpackb(result) == {
        'endpoint': 'v0/slave_send',
        'data': {
            'ids': [1, 2, 3, 4]
        }
    }


def test_quotas(mocker):
    from iris.sender.quota import ApplicationQuota
    from iris.metrics import stats
    from gevent import sleep
    mocker.patch('iris.sender.quota.ApplicationQuota.get_new_rules',
                 return_value=[(
                     'testapp', 5, 2, 120, 120, 'testuser',
                     'user', 'iris-plan', 10
                 )])
    mocker.patch('iris.sender.quota.ApplicationQuota.notify_incident')
    mocker.patch('iris.sender.quota.ApplicationQuota.notify_target')
    quotas = ApplicationQuota(None, None, None, None)
    sleep(1)

    # ensure drop messages don't count
    assert quotas.allow_send({'application': 'testapp', 'mode': 'drop'})
    assert quotas.allow_send({'application': 'testapp', 'mode': 'drop'})
    assert quotas.allow_send({'application': 'testapp', 'mode': 'drop'})
    assert quotas.allow_send({'application': 'testapp', 'mode': 'drop'})
    assert quotas.allow_send({'application': 'testapp', 'mode': 'drop'})
    assert quotas.allow_send({'application': 'testapp', 'mode': 'drop'})

    assert stats['quota_soft_exceed_cnt'] == 0
    assert stats['quota_hard_exceed_cnt'] == 0

    # but normal ones do
    assert quotas.allow_send({'application': 'testapp'})
    assert quotas.allow_send({'application': 'testapp'})
    assert quotas.allow_send({'application': 'testapp'})  # Breach soft quota
    assert quotas.allow_send({'application': 'testapp'})
    assert quotas.allow_send({'application': 'testapp'})
    assert not quotas.allow_send({'application': 'testapp'})  # Breach hard quota
    assert not quotas.allow_send({'application': 'testapp'})
    assert stats['quota_soft_exceed_cnt'] == 3
    assert stats['quota_hard_exceed_cnt'] == 2

    assert stats['app_testapp_quota_hard_usage_pct'] == 100
    assert stats['app_testapp_quota_soft_usage_pct'] == 200

    for _ in range(10):
        assert quotas.allow_send({'application': 'app_without_quota'})


def test_aggregate_audit_msg(mocker):
    mock_iris_client = mocker.patch('iris.sender.cache.iris_client')
    mock_iris_client.get.return_value.json.return_value = fake_plan
    from iris.bin.sender import (
        fetch_and_prepare_message, message_queue, per_mode_send_queues,
        plan_aggregate_windows
    )

    send_queue = per_mode_send_queues.setdefault('email', gevent.queue.Queue())

    init_queue_with_item(message_queue, fake_message)
    init_queue_with_item(send_queue)

    now = time.time()
    msg_aggregate_key = (
        fake_message['plan_id'],
        fake_message['application'],
        fake_message['priority'],
        fake_message['target'])
    from collections import defaultdict
    plan_aggregate_windows[msg_aggregate_key] = defaultdict(int)
    plan_aggregate_windows[msg_aggregate_key][now] = 10
    plan_aggregate_windows[msg_aggregate_key][now - 60] = 10

    mocker.patch('iris.bin.sender.cache').plans = {fake_plan['id']: fake_plan}

    mocker.patch('iris.bin.sender.spawn')
    from iris.bin.sender import spawn as mock_spawn

    # run code to test
    fetch_and_prepare_message()

    # examine results
    assert send_queue.qsize() == 0
    from iris.sender import auditlog
    mock_spawn.assert_called_with(
        auditlog.message_change,
        fake_message['message_id'],
        auditlog.SENT_CHANGE,
        '',
        '',
        "Aggregated with key (19546, test-app, high, test-user)")


def test_handle_api_notification_request_invalid_message(mocker):
    mocker.patch('iris.bin.sender.set_target_contact').return_value = True
    mocker.patch('iris.metrics.stats')

    from iris.bin.sender import message_send_enqueue
    from iris.sender.rpc import handle_api_notification_request, send_funcs
    from iris.sender.shared import per_mode_send_queues

    send_funcs['message_send_enqueue'] = message_send_enqueue
    send_queue = per_mode_send_queues.setdefault('email', gevent.queue.Queue())

    mock_socket = mocker.MagicMock()
    handle_api_notification_request(mock_socket, mocker.MagicMock(), {
        'data': {
            'application': 'test_app',
        }
    })
    mock_socket.sendall.assert_called_with(msgpack.packb('INVALID role'))

    handle_api_notification_request(mock_socket, mocker.MagicMock(), {
        'data': {
            'role': 'user',
            'application': 'test_app',
        }
    })
    mock_socket.sendall.assert_called_with(msgpack.packb('INVALID target'))

    mocker.patch('iris.sender.rpc.cache').targets_for_role.return_value = ['foo']
    handle_api_notification_request(mock_socket, mocker.MagicMock(), {
        'data': {
            'target': 'foo',
            'role': 'user',
            'application': 'test_app',
        }
    })
    mock_socket.sendall.assert_called_with(msgpack.packb('INVALID body'))

    handle_api_notification_request(mock_socket, mocker.MagicMock(), {
        'data': {
            'target': 'foo',
            'role': 'user',
            'application': 'test_app',
            'template': 'test',
        }
    })
    mock_socket.sendall.assert_called_with(msgpack.packb('INVALID context'))

    # should work when user is setting body key
    handle_api_notification_request(mock_socket, mocker.MagicMock(), {
        'data': {
            'target': 'foo',
            'role': 'user',
            'application': 'test_app',
            'body': 'test',
        }
    })
    mock_socket.sendall.assert_called_with(msgpack.packb('OK'))

    # should work when user is setting template key
    handle_api_notification_request(mock_socket, mocker.MagicMock(), {
        'data': {
            'target': 'foo',
            'role': 'user',
            'application': 'test_app',
            'template': 'test',
            'context': {},
        }
    })
    mock_socket.sendall.assert_called_with(msgpack.packb('OK'))

    # should work when user is setting email_html key
    handle_api_notification_request(mock_socket, mocker.MagicMock(), {
        'data': {
            'target': 'foo',
            'role': 'user',
            'application': 'test_app',
            'email_html': '<p>test</p>',
        }
    })
    mock_socket.sendall.assert_called_with(msgpack.packb('OK'))

    # drain out send queue
    while send_queue.qsize() > 0:
        send_queue.get()


def test_sanitize_unicode_dict():
    from jinja2.sandbox import SandboxedEnvironment
    from iris.utils import sanitize_unicode_dict

    # Use jinja the same way as in sender
    env = SandboxedEnvironment(autoescape=False)
    template = env.from_string('{{var}} {{var2}} {{nested.nest}}')
    bad_context = {'var': b'\xe2\x80\x99', 'var2': 2, 'nested': {'nest': b'\xe2\x80\x99'}}

    assert '\\xe2' in template.render(**bad_context)

    good_render = template.render(**sanitize_unicode_dict(bad_context))
    assert '\\xe2' not in good_render
    assert b'\xe2\x80\x99'.decode('utf-8') in good_render



import datetime
import logging
import ujson
from falcon import HTTP_201, HTTPBadRequest, HTTPNotFound

from iris import db

logger = logging.getLogger(__name__)


class alertmanager(object):
    allow_read_no_auth = False

    def validate_post(self, body):
        if not all(k in body for k in("version", "status", "alerts")):
            raise HTTPBadRequest('missing version, status and/or alert attributes')

        if 'iris_plan' not in body["groupLabels"]:
            raise HTTPBadRequest('missing iris_plan in group labels')

    def create_context(self, body):
        context_json_str = ujson.dumps(body)
        if len(context_json_str) > 65535:
            logger.warning('POST to alertmanager exceeded acceptable size')
            raise HTTPBadRequest('Context too long')

        return context_json_str

    def on_post(self, req, resp):
        '''
        This endpoint is compatible with the webhook post from Alertmanager.
        Simply configure alertmanager with a receiver pointing to iris, like
        so:

        receivers:
        - name: 'iris-team1'
          webhook_configs:
            - url: http://iris:16649/v0/webhooks/alertmanager?application=test-app&key=sdffdssdf

        Where application points to an application and key in Iris.

        For every POST from alertmanager, a new incident will be created, if the iris_plan label
        is attached to an alert.
        '''
        alert_params = ujson.loads(req.context['body'])
        self.validate_post(alert_params)

        with db.guarded_session() as session:
            plan = alert_params['groupLabels']['iris_plan']
            plan_id = session.execute('SELECT `plan_id` FROM `plan_active` WHERE `name` = :plan',
                                      {'plan': plan}).scalar()
            if not plan_id:
                raise HTTPNotFound()

            app = req.context['app']

            context_json_str = self.create_context(alert_params)

            app_template_count = session.execute('''
                SELECT EXISTS (
                  SELECT 1 FROM
                  `plan_notification`
                  JOIN `template` ON `template`.`name` = `plan_notification`.`template`
                  JOIN `template_content` ON `template_content`.`template_id` = `template`.`id`
                  WHERE `plan_notification`.`plan_id` = :plan_id
                  AND `template_content`.`application_id` = :app_id
                )
            ''', {'app_id': app['id'], 'plan_id': plan_id}).scalar()

            if not app_template_count:
                logger.warning('no plan template exists for this app')
                raise HTTPBadRequest('No plan template actions exist for this app')

            data = {
                'plan_id': plan_id,
                'created': datetime.datetime.utcnow(),
                'application_id': app['id'],
                'context': context_json_str,
                'current_step': 0,
                'active': True,
            }

            incident_id = session.execute(
                '''INSERT INTO `incident` (`plan_id`, `created`, `context`,
                                           `current_step`, `active`, `application_id`)
                   VALUES (:plan_id, :created, :context, 0, :active, :application_id)''',
                data).lastrowid

            session.commit()
            session.close()

        resp.status = HTTP_201
        resp.set_header('Location', '/incidents/%s' % incident_id)
        resp.body = ujson.dumps(incident_id)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
logger = logging.getLogger(__name__)


class dummy(object):
    def __init__(self, config):
        pass

    def get(self, role, target):
        if role == '_invalid_role' or target == '_invalid_user':
            logger.warning('Deliberately returning 0 results for %s:%s', role, target)
            return []
        return ['foo']



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# -*- coding:utf-8 -*-

from .. import utils
import logging
logger = logging.getLogger(__name__)

_registered_plugins = []
_plugins = None


def parse_response(msg):
    args = msg.split(' ')
    cmd = args.pop(0).lower()
    return cmd, args


class IrisPlugin(object):
    name = 'IrisDefault'
    phone_response_menu = {
        '2': {
            'title': 'Press 2 to claim.',
            'cmd': 'claim',
        },
    }

    def __init__(self, config):
        self.config = config

    @classmethod
    def get_phone_menu_text(cls):
        if cls.phone_response_menu:
            return ' '.join([item['title'] for (_, item) in
                            cls.phone_response_menu.items()])
        else:
            return ''

    def handle_response(self, mode, msg_id, source, content, batch=False):
        if mode == 'call':
            digits = content
            logger.info('%s Plugin got phone call response for message(%s): %s:',
                        self.name, msg_id, digits)
            if digits not in self.phone_response_menu:
                msg = 'Got unknown option from user: ' + digits
                logger.warning(msg)
                return msg
            cmd, args = self.phone_response_menu[digits]['cmd'], None
        else:
            logger.info('%s Plugin got %s response for message(%s): %s',
                        self.name, mode, msg_id, content)
            cmd, args = parse_response(content)
        if batch:
            cmd = 'batch_' + cmd
        return self.process_command(msg_id, source, mode, cmd, args)

    def process_iris_claim(self, msg_id, source, mode, cmd, args=None):
        iid = utils.get_incident_id_from_message_id(msg_id)
        if not iid:
            return 'Failed to match incident for message(%s).' % msg_id
        owner = utils.lookup_username_from_contact(mode, source)
        if not owner:
            return 'Failed to identify owner for the incident.'
        is_active, previous_owner = utils.claim_incident(iid, owner)
        if is_active:
            logger.info(('Failed to claim incident(%s) for message: %s. '
                         'owner: %s, mode: %s, args: %s'),
                        iid, msg_id, owner, mode, args)
            return 'Failed to claim incident for message: %s.' % str(msg_id)
        else:
            if previous_owner:
                return 'Iris incident(%s) claimed, previously claimed by %s.' % (iid, previous_owner)
            else:
                return 'Iris incident(%s) claimed.' % iid

    def process_iris_batch_claim(self, msg_id, source, mode, cmd, args=None):
        owner = utils.lookup_username_from_contact(mode, source)
        if not owner:
            return 'Failed to identify owner for the incident.'
        utils.claim_incidents_from_batch_id(msg_id, owner)
        return 'All iris incidents claimed for batch id %s.' % msg_id

    def process_claim_all(self, msg_ids, source, mode):
        owner = utils.lookup_username_from_contact(mode, source)
        if not owner:
            return 'Failed to identify owner for these incidents.'
        if not msg_ids:
            return 'No messages to claim.'
        incident_ids = utils.get_incident_ids_from_message_ids(msg_ids)
        if not incident_ids:
            return 'No messages to claim.'

        claimed, not_claimed = utils.claim_bulk_incidents(incident_ids, owner)

        msg = []
        if claimed:
            msg.append('Iris Incidents claimed (%s): %s' % (len(claimed), ', '.join(map(str, claimed))))
        if not_claimed:
            msg.append('Iris Incidents failed to claim (%s): %s' % (len(not_claimed), ', '.join(map(str, not_claimed))))

        if msg:
            return '\n'.join(msg)
        else:
            return 'Unknown claim all result'

    def process_command(self, msg_id, source, mode, cmd, args=None):
        if cmd == 'claim':
            return self.process_iris_claim(msg_id, source, mode, cmd, args)
        elif cmd == 'batch_claim':
            return self.process_iris_batch_claim(msg_id, source, mode, cmd, args)
        elif cmd == 'claim_all':
            return self.process_claim_all(msg_id, source, mode)
        else:
            return 'Unknown command.'


def init_plugins(config):
    global _plugins
    if not _plugins:
        _plugins = {}
        for plugin in _registered_plugins:
            name = plugin.name
            try:
                _plugins[name] = plugin(config.get(name, {}))
            except Exception:
                logger.exception('failed to register plugin: %s', name)
            else:
                logger.info('registered plugin: %s', name)
        _plugins[IrisPlugin.name] = IrisPlugin({})
    return _plugins


def register_plugin():
    def wrapper(plugin):
        if not issubclass(plugin, IrisPlugin):
            raise ValueError('%s is not a subclass of %s' %
                             (type(plugin), type(IrisPlugin)))
        _registered_plugins.append(plugin)
        return plugin
    return wrapper


def find_plugin(plugin_name, include_deafault=True):
    plugin = _plugins.get(plugin_name)
    if plugin:
        return plugin
    elif include_deafault:
        return _plugins[IrisPlugin.name]


# core.py itself is excluded from __init__.py
from . import *  # noqa



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from gevent.pywsgi import WSGIServer
from iris.api import API
import sys
import logging
import yaml

if __name__ == '__main__':
    config_path = sys.argv[1]
    with open(config_path, 'r') as config_file:
        config = yaml.safe_load(config_file)

    app = API(config)
    logging.basicConfig()
    server = config['server']
    print('LISTENING: %(host)s:%(port)d' % server)
    WSGIServer((server['host'], server['port']), app).serve_forever()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris.custom_import import import_custom_module
import logging
logger = logging.getLogger(__name__)


class IrisRoleLookupException(Exception):
    pass


def get_role_lookups(config):
    modules = config.get('role_lookups', [])

    # default to only support user and mailing_list.
    if not modules:
        modules = ['user', 'mailing_list']

    imported_modules = []
    for m in modules:
        try:
            imported_modules.append(
                import_custom_module('iris.role_lookup', m)(config))
            logger.info('Loaded lookup modules: %s', m)
        except Exception:
            logger.exception('Failed to load role lookup module: %s', m)

    return imported_modules
